import os
import errno
def create_symlink(source, target_link):
    """
    """
    try:
        os.symlink(source, target_link)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
            os.remove(target_link)
            os.symlink(source, target_link)
        else:
            raise e


import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from pyearth.system.define_global_variables import *
from pyearth.visual.color.create_diverge_rgb_color_hex import create_diverge_rgb_color_hex

mpl.rcParams['hatch.linewidth'] = 0.5


def boxplot_data(aData_in,
                 aLabel_x_in,
                 aLabel_y_in,
                 sFilename_out,
                 iDPI_in=None,
                 ncolumn_in=None,
                 iSize_x_in=None,
                 iSize_y_in=None,
                 dMax_y_in=None,
                 dMin_y_in=None,
                 dSpace_y_in=None,
                 aMarker_in=None,
                 aColor_in=None,
                 aHatch_in=None,
                 sLabel_y_in=None,
                 aLabel_legend_in=None,
                 aLocation_legend_in=None,
                 sFormat_y_in=None,
                 sLocation_legend_in=None,
                 sTitle_in=None):
    """Draw a boxplot

    Args:
        aData_in (_type_): _description_
        aLabel_x_in (_type_): _description_
        aLabel_y_in (_type_): _description_
        sFilename_out (_type_): _description_
        iDPI_in (_type_, optional): _description_. Defaults to None.
        ncolumn_in (_type_, optional): _description_. Defaults to None.
        iSize_x_in (_type_, optional): _description_. Defaults to None.
        iSize_y_in (_type_, optional): _description_. Defaults to None.
        dMax_y_in (_type_, optional): _description_. Defaults to None.
        dMin_y_in (_type_, optional): _description_. Defaults to None.
        dSpace_y_in (_type_, optional): _description_. Defaults to None.
        aMarker_in (_type_, optional): _description_. Defaults to None.
        aColor_in (_type_, optional): _description_. Defaults to None.
        aHatch_in (_type_, optional): _description_. Defaults to None.
        sLabel_y_in (_type_, optional): _description_. Defaults to None.
        aLabel_legend_in (_type_, optional): _description_. Defaults to None.
        aLocation_legend_in (_type_, optional): _description_. Defaults to None.
        sFormat_y_in (_type_, optional): _description_. Defaults to None.
        sLocation_legend_in (_type_, optional): _description_. Defaults to None.
        sTitle_in (_type_, optional): _description_. Defaults to None.
    """
    # if a data does not have a consistent dimension, can we use boxplot?
    # Yes, in this case, a list of list is a better data structure for the data

    nData = len(aData_in)

    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9

    if dMax_y_in is not None:
        dMax_y = dMax_y_in
    else:
        dMax_y = -9999
        for i in range(nData):
            aData_dummy = aData_in[i]
            for j in range(len(aData_dummy)):
                dMax_y_dummy = np.max(aData_dummy[j])
                dMax_y = np.max([dMax_y, dMax_y_dummy])

    if dMin_y_in is not None:
        dMin_y = dMin_y_in
    else:
        dMin_y = 9999
        for i in range(nData):
            aData_dummy = aData_in[i]
            for j in range(len(aData_dummy)):
                dMin_y_dummy = np.min(aData_dummy[j])
                dMin_y = np.min([dMin_y, dMin_y_dummy])

    if (dMax_y <= dMin_y):
        return

    if sLabel_y_in is not None:
        sLabel_y = sLabel_y_in
    else:
        sLabel_y = ''

    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    if sFormat_y_in is not None:
        iFlag_format_y = 1
        sFormat_y = sFormat_y_in
    else:
        iFlag_format_y = 0

    if sLocation_legend_in is not None:
        sLocation_legend = sLocation_legend_in
    else:
        sLocation_legend = "upper right"

    if aLocation_legend_in is not None:
        aLocation_legend = aLocation_legend_in
    else:
        aLocation_legend = (1.0, 1.0)

    if ncolumn_in is not None:
        ncolumn = ncolumn_in
    else:
        ncolumn = 1

    if aColor_in is not None:
        aColor = aColor_in
    else:
        if (nData >= 3):
            aColor = create_diverge_rgb_color_hex(nData)
        else:
            if nData == 2:
                aColor = ['red', 'blue']
            else:
                aColor = ['red']

    if aHatch_in is not None:
        aHatch = aHatch_in
    else:
        aHatch = np.full(nData, '+')

    fig = plt.figure(dpi=iDPI)
    fig.set_figwidth(iSize_x)
    fig.set_figheight(iSize_y)
    ax = fig.add_axes([0.1, 0.5, 0.8, 0.4])

    dWidth_total = 0.7

    aLegend_artist = []
    aLabel = []

    for i in np.arange(1, nData + 1, 1):

        data1 = aData_in[i-1]

        # the data1 is still considered a list
        nData1 = len(data1)
        # plot group

        # set x tick
        x_center = i-1
        # set width
        dWidth = 0.8 * dWidth_total / nData1

        for j in range(1, nData1+1, 1):
            # plot individual data
            data_raw = data1[j-1]

            q1_raw = np.percentile(data_raw, 25)
            q3_raw = np.percentile(data_raw, 75)
            qm_raw = np.median(data_raw)
            #qm = qm_raw  # np.log10(qm_raw)

            iqr_raw = q3_raw - q1_raw
            #iqr_log = np.log10(iqr_raw)

            q0_raw = np.percentile(data_raw, 0.35)
            q4_raw = np.percentile(data_raw, 99.65)  # q3_raw + 1.5 * iqr_raw

            # set
            x_left = x_center - 0.5 * dWidth_total + (j-1) * dWidth
            x_right = x_left + dWidth
            x_mid = 0.5*(x_left + x_right)

            y_bot = q1_raw
            y_top = q3_raw
            height = y_top-y_bot
            # draw box
            rect = mpl.patches.Rectangle((x_left, y_bot), dWidth, height, linewidth=0.5,
                                     facecolor=aColor[j-1], edgecolor='k')  # , hatch = aHatch[j-1 ]

            if i == 1:
                aLegend_artist.append(rect)
                aLabel.append(aLabel_legend_in[j-1])

            ax.add_patch(rect)
            # draw lower q1
            x0 = [x_mid, x_mid]
            y0 = [q0_raw, q1_raw]
            line, = ax.plot(x0, y0,
                            color='black', linestyle='--', linewidth=1)
            # draw short line
            x0 = [x_mid - 0.2 * dWidth, x_mid + 0.2 * dWidth]
            y0 = [q0_raw, q0_raw]
            line, = ax.plot(x0, y0,
                            color='black', linestyle='-', linewidth=1)

            # draw upper q3
            x0 = [x_mid, x_mid]
            y0 = [q3_raw, q4_raw]
            line, = ax.plot(x0, y0,
                            color='black', linestyle='--', linewidth=1)

            # draw short line
            x0 = [x_mid - 0.2 * dWidth,  x_mid + 0.2 * dWidth]
            y0 = [q4_raw, q4_raw]
            line, = ax.plot(x0, y0,
                            color='black', linestyle='-', linewidth=1)

            # draw median
            x0 = [x_mid - 0.5 * dWidth,  x_mid + 0.5 * dWidth]
            y0 = [qm_raw, qm_raw]
            line, = ax.plot(x0, y0,
                            color='black', linestyle='-', linewidth=1)

            # whiskers outlier
            dummy_index = np.where(data_raw < q0_raw)[0]
            for k in range(len(dummy_index)):
                x0 = [x_mid]
                y0 = [data_raw[dummy_index[k]]]
                ax.plot(x0, y0, marker="o", markersize=3,
                        markeredgecolor="black", markerfacecolor="white")
                pass
            dummy_index = np.where(data_raw > q4_raw)[0]
            for k in range(len(dummy_index)):
                x0 = [x_mid]
                y0 = [data_raw[dummy_index[k]]]
                ax.plot(x0, y0, marker="o", markersize=3,
                        markeredgecolor="black", markerfacecolor="white")
                pass

        pass

    x = np.arange(len(aLabel_x_in))
    # Add some text for labels, title and custom x-axis tick labels, etc.
    ax.set_ylabel(sLabel_y)
    ax.set_title(sTitle)
    ax.set_xticks(x)
    ax.set_xticklabels(aLabel_x_in)

    if (iFlag_format_y == 1):
        ax.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter(sFormat_y))

    ax.set_xlim(-1, nData)
    ax.set_ylim(dMin_y, dMax_y)
    ax.grid(which='major', color='grey', linestyle='--', axis='y')
    #ax.set_yscale('log')

    ax.legend(aLegend_artist, aLabel, bbox_to_anchor=aLocation_legend,
              loc=sLocation_legend, fontsize=12, ncol=ncolumn)

    plt.savefig(sFilename_out, bbox_inches='tight')
    plt.close('all')
    plt.clf()



import numpy as np
def kde_support(aData_in,
                bw_in,
                iGridsize_in,
                cut_in,
                aClip_in):
    """
    Establish support for a kernel density estimate.

    Args:
        aData_in (numpy.array): Input data
        bw_in (string): bandwidth
        iGridsize_in (int): Gridsize
        cut_in (float): Factor
        aClip_in (numpy.array): Bound

    Returns:
        numpy.array: kde
    """
    support_min = max(aData_in.min() - bw_in * cut_in, aClip_in[0])
    support_max = min(aData_in.max() + bw_in * cut_in, aClip_in[1])
    return np.linspace(support_min, support_max, iGridsize_in)


def scipy_bivariate_kde(aX_in,
                        aY_in,
                        bw_in,
                        iGridsize_in,
                        cut_in,
                        aClip_in):
    """
    Compute a bivariate kde using scipy.

    Args:
        aX_in (numpy.array): X array    
        aY_in (numpy.array): Y array
        bw_in (string): bandwidth
        iGridsize_in (int): Gridsize
        cut_in (float): Factor
        aClip_in (numpy.array): Bound

    Raises:


    Returns:
        numpy.array: kde
    """

    try:
        import scipy
    except ImportError as e:
        raise ImportError(
            "The package 'scipy' is required for this function to run.") from e

    data = np.c_[aX_in, aY_in]
    kde = scipy.stats.gaussian_kde(data.T, bw_method=bw_in)
    data_std = data.std(axis=0, ddof=1)
    if isinstance(bw_in, str):
        bw_in = "scotts" if bw_in == "scott" else bw_in
        bw_x = getattr(kde, "%s_factor" % bw_in)() * data_std[0]
        bw_y = getattr(kde, "%s_factor" % bw_in)() * data_std[1]
    elif np.isscalar(bw_in):
        bw_x, bw_y = bw_in, bw_in
    else:
        msg = ("Cannot specify a different bandwidth for each dimension "
               "with the scipy backend. You should install statsmodels.")
        raise ValueError(msg)

    x_support = kde_support(
        data[:, 0], bw_x, iGridsize_in, cut_in, aClip_in[0])
    y_support = kde_support(
        data[:, 1], bw_y, iGridsize_in, cut_in, aClip_in[1])
    xx, yy = np.meshgrid(x_support, y_support)
    z = kde([xx.ravel(), yy.ravel()]).reshape(xx.shape)
    return xx, yy, z



from datetime import datetime
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from pyearth.system.define_global_variables import *


def plot_time_series_data_with_two_y_axis(aTime_all, aData_all,
                                          sFilename_out,
                                          iDPI_in=None,
                                          iFlag_trend_in=None,
                                          iReverse_y_in=None,
                                          iSize_x_in=None,
                                          iSize_y_in=None,
                                          aMax_y_in=None,
                                          aMin_y_in=None,
                                          aSpace_y_in=None,
                                          aMarker_in=None,
                                          aColor_in=None,
                                          aLinestyle_in=None,
                                          aLabel_y_in=None,
                                          aLabel_legend_in=None,
                                          sTitle_in=None):
    nData = len(aData_all)
    aData = [aData_all]
    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_trend_in is not None:
        iFlag_trend = 1
    else:
        iFlag_trend = 0

    if iReverse_y_in is not None:
        iReverse_y = 1
    else:
        iReverse_y = 0

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12
    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9

    if aLabel_y_in is not None:
        sLabel_y1 = aLabel_y_in[0]
        sLabel_y2 = aLabel_y_in[1]
    else:
        sLabel_y1 = ''
        sLabel_y2 = ''
    if aLabel_legend_in is not None:
        aLabel_legend = aLabel_legend_in
        sLabel_legend1 = aLabel_legend_in[0]
        sLabel_legend2 = aLabel_legend_in[1]
    else:
        sLabel_legend1 = ''
        sLabel_legend2 = ''
    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    if aMarker_in is not None:
        aMarker = aMarker_in
    else:
        aMarker = np.full(nData, '+')

    if aColor_in is not None:
        aColor = aColor_in
    else:
        aColor = np.full(nData, 'black')

    if aLinestyle_in is not None:
        aLinestyle = aLinestyle_in
    else:
        aLinestyle = np.full(nData, '-')

    if aMax_y_in is not None:
        dMax_y1 = aMax_y_in[0]
        dMax_y2 = aMax_y_in[1]
    else:
        dMax_y1 = np.nanmax(aData[0]) * 1.2
        dMax_y2 = np.nanmax(aData[1]) * 1.2
    if aMin_y_in is not None:
        dMin_y1 = aMin_y_in[0]
        dMin_y2 = aMin_y_in[1]
    else:
        # dMin_Y = np.nanmin(aData) #if it has negative value, change here
        dMin_y1 = np.nanmin(aData[0])
        dMin_y2 = np.nanmin(aData[1])

    # if (dMax_Y <= dMin_Y ):
    #    return

    if aSpace_y_in is not None:
        dSpace_y1 = aSpace_y_in[0]
        dSpace_y2 = aSpace_y_in[1]
    else:
        pass

    fig = plt.figure(dpi=iDPI)
    fig.set_figwidth(iSize_x)
    fig.set_figheight(iSize_y)
    ax1 = fig.add_axes([0.1, 0.5, 0.8, 0.4])
    pYear = mpl.dates.YearLocator(1)   # every year
    pMonth = mpl.dates.MonthLocator()  # every month
    sYear_format = mpl.dates.DateFormatter('%Y')
    # plot the first axis
    i = 1
    x1 = aTime_all[i-1]

    y1 = aData_all[i-1]
    ax1.plot(x1, y1,
             color=aColor[i-1], linestyle=aLinestyle[i-1],
             marker=aMarker[i-1], markersize=0.5,
             label=aLabel_legend[i-1])

    # calculate linear regression
    nan_index = np.where(y1 == missing_value)
    y1[nan_index] = np.nan
    good_index = np.where(~np.isnan(y1))
    if iFlag_trend == 1:
        x_dummy = np.array([i.timestamp() for i in x1])
        x_dummy = x_dummy[good_index]
        y_dummy = y1[good_index]
        coef = np.polyfit(x_dummy, y_dummy, 1)
        poly1d_fn = np.poly1d(coef)
        mn = np.min(x_dummy)
        mx = np.max(x_dummy)
        x2 = [mn, mx]
        y2 = poly1d_fn(x2)
        x2 = [datetime.fromtimestamp(i) for i in x2]
        ax1.plot(x2, y2, color='orange', linestyle='-.',  linewidth=0.5)
    # plot the second axis
    ax2 = ax1.twinx()
    i = 2
    x2 = aTime_all[i-1]
    y2 = aData_all[i-1]
    ax2.plot(x2, y2,
             color=aColor[i-1], linestyle=aLinestyle[i-1],
             marker=aMarker[i-1],
             label=aLabel_legend[i-1])

    ax1.axis('on')
    ax1.grid(which='major', color='grey', linestyle='--', axis='y')

    # ax1.set_aspect(dRatio)  #this one set the y / x ratio
    ax1.xaxis.set_major_locator(pYear)
    ax1.xaxis.set_minor_locator(pMonth)
    ax1.xaxis.set_major_formatter(sYear_format)
    ax1.tick_params(axis="x", labelsize=10)
    ax1.tick_params(axis="y", labelsize=10)

    ax1.set_xmargin(0.05)
    ax1.set_ymargin(0.15)

    ax1.set_xlabel('Year', fontsize=12)
    ax1.set_ylabel(sLabel_y1, fontsize=12)
    ax2.set_ylabel(sLabel_y2, fontsize=12)
    ax1.set_title(sTitle, loc='center', fontsize=15)
    # round to nearest years...
    x_min = np.datetime64(aTime_all[1][0], 'Y')
    x_max = np.datetime64(aTime_all[1][-1], 'Y') + np.timedelta64(1, 'Y')
    ax1.set_xlim(x_min, x_max)
    if dMax_y1 < 1000 and dMax_y1 > 0.001:
        ax1.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
    else:
        ax1.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1e'))

    ax1.yaxis.set_major_locator(mpl.ticker.MultipleLocator(dSpace_y1))
    ax2.yaxis.set_major_locator(mpl.ticker.MultipleLocator(dSpace_y2))
    if (iReverse_y == 1):
        ax1.set_ylim(dMax_y1, dMin_y1)
    else:
        ax1.set_ylim(dMin_y1, dMax_y1)

    ax1.legend(bbox_to_anchor=(1.0, 1.0), loc="upper right", fontsize=12)
    ax2.legend(bbox_to_anchor=(1.0, 1.0), loc="upper right", fontsize=12)
    print(sFilename_out)
    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')
    plt.clf()
    # print('finished plotting')



import os
def replace_variable_in_netcdf(sFilename_old, sFilename_new, aData_in, aVariable_in):
    try:
        import netCDF4 as nc
    except ImportError as e:
        raise ImportError("The package 'netCDF4' is required for this function to run.") from e

    if os.path.exists(sFilename_old):
        print("Yep, I can read that file: "+ sFilename_old)
    else:
        print("Nope, the path doesn't reach your file. Go research filepath in python")
        exit
    pDatasets_in = nc.Dataset(sFilename_old)
    netcdf_format = pDatasets_in.file_format
    #output file
    pDatasets_out = nc.Dataset(sFilename_new, "w", format=netcdf_format)
    for sKey, iValue in pDatasets_in.dimensions.items():
        dummy = len(iValue)
        if not iValue.isunlimited():            
            pDatasets_out.createDimension(sKey, dummy)            
        else:
            pDatasets_out.createDimension(sKey, dummy )

    #copy global attribute
    
    for attr in pDatasets_in.ncattrs():
        #print(':::GlobalAtt:', attr,' Val:', getattr(pDatasets_in,attr))
        pDatasets_out.setncattr( attr, pDatasets_in.getncattr( attr ) )

    # Copy variables
    aAttribute = list()
    aDimension = list()
    aDataType = list()
    for sKey, aValue in pDatasets_in.variables.items():       
      
        # we need to take care of rec dimension
 
        if(sKey in aVariable_in): #only copy other dataset first
            aDataType.append(aValue.datatype)
            aDimension.append(aValue.dimensions)
            #pUnit = aValue.units
            for sAttribute in aValue.ncattrs():                
                if sAttribute == '_FillValue' or sAttribute =='missing_value':
                    pass
                else:
                    aAttribute.append(sAttribute)
            pass
        else:

            outVar = pDatasets_out.createVariable(sKey, aValue.datatype, aValue.dimensions )        
            for sAttribute in aValue.ncattrs():            
                outVar.setncatts( { sAttribute: aValue.getncattr(sAttribute) } )

            outVar[:] = aValue[:]
            
    #close the output file
    #replace variable 
    for i in range(len(aVariable_in)):
        sVariable_in = aVariable_in[i]
        aData_in0 = aData_in[i]
        pVar3 = pDatasets_out.createVariable(sVariable_in, aDataType[i], aDimension[i], fill_value=-9999  ) 
        pVar3[:] = aData_in0
        pValue = pDatasets_in.variables[sVariable_in]
        for sAttribute in pValue.ncattrs():            
            if sAttribute == '_FillValue' or sAttribute =='missing_value':
                pass
            else:
                pVar3.setncatts( { sAttribute: pValue.getncattr(sAttribute) } )
        pVar3.missing_value = -9999 
    
    
    #pVar3.description = sVariable_in
    #pVar3.units = pUnit
    pDatasets_out.close()
    return



from collections import OrderedDict
import numpy as np
def create_line_style(nStyle):
    linestyles = OrderedDict(
        [('solid',               (0, ())),
         ('loosely dotted',      (0, (1, 10))),
            ('dotted',              (0, (1, 5))),
            ('densely dotted',      (0, (1, 1))),

            ('loosely dashed',      (0, (5, 10))),
            ('dashed',              (0, (5, 5))),
            ('densely dashed',      (0, (5, 1))),

            ('loosely dashdotted',  (0, (3, 10, 1, 10))),
            ('dashdotted',          (0, (3, 5, 1, 5))),
            ('densely dashdotted',  (0, (3, 1, 1, 1))),

            ('loosely dashdotdotted', (0, (3, 10, 1, 10, 1, 10))),
            ('dashdotdotted',         (0, (3, 5, 1, 5, 1, 5))),
            ('densely dashdotdotted', (0, (3, 1, 1, 1, 1, 1)))])

    index = np.arange(nStyle)

    aStyle = list()
    aLine_style_out = list()
    for key, value in linestyles.items():
        aStyle.append(value)

    for i in range(nStyle):
        aLine_style_out.append(aStyle[i])
    return aLine_style_out



def find_index_by_longitude_latitude(dLon_in,
                         dLat_in,
                         dLon_min_in,
                         dLat_max_in,
                         dResolution_x_in,
                         dResolution_y_in,
                         iFlag_center_in=None):
    """
    Find the index of a lcoation in a large 2D array, we requre the location is at center, the longitude should be in the same range (-180, 180) or (0, 360)

    Args:
        dLon_in (float): _description_
        dLat_in (float): _description_
        dLon_min_in (float): _description_
        dLat_max_in (float): _description_
        dResolution_x_in (float): The mesh cell resolution in x direction
        dResolution_y_in (float): The mesh cell resolution in y directioniption_
        iFlag_center_in (int, optional): Whether the max/min is at center. Defaults to None.
    """
    if iFlag_center_in is None:
        iFlag_center = 1
    else:
        iFlag_center = iFlag_center_in

    if iFlag_center == 1:  # when the max and min are center
        iRow_index = round((dLat_max_in - dLat_in) / dResolution_y_in)
        iColumn_index = round((dLon_in - (dLon_min_in)) / dResolution_x_in)
        pass
    else:  # when the max and min are upper corner
        iRow_index = round(
            ((dLat_max_in - 0.5*dResolution_y_in) - dLat_in) / dResolution_y_in)
        iColumn_index = round(
            (dLon_in - (dLon_min_in + 0.5*dResolution_x_in)) / dResolution_x_in)
        pass

    return iRow_index, iColumn_index



import os
def prepare_parafly_slurm_job_script(sBasename_job,
                                     sBasename_parafly,
                                     sDirectory_job,
                                     sEmail,
                                     iWalltime_in=None,
                                     nNode_in=None,
                                     nThread_in=None,
                                     sJob_name_in=None,
                                     sPython_env_in=None,
                                     sQueue_in=None):
    """
    prepare a job script for parafly

    Args:
        sBasename_job (string): The base job name
        sBasename_parafly (string): The parafly script filename
        sDirectory_job (string): The path where parafly runs the script
        sEmail (string): Email address for job notification
        iWalltime_in (int, optional): Walltime in hour. Defaults to None.
        nNode_in (int, optional): NUmber of node. Defaults to None.
        nThread_in (int, optional): Number of node. Defaults to None.
        sJob_name_in (string, optional): Job name. Defaults to None.
        sPython_env_in (string, optional): Conda python environment. Defaults to None.
        sQueue_in (string, optional): HPC queue or partition. Defaults to None.
    """
    if iWalltime_in is not None:
        iWalltime = iWalltime_in
    else:
        iWalltime = 2
    if nNode_in is not None:
        iNode = nNode_in
    else:
        iNode = 1
    if nThread_in is not None:
        nThread = nThread_in
    else:
        nThread = 40

    if sJob_name_in is not None:
        sJob_name = sJob_name_in
    else:
        sJob_name = 'parafly'
    if sPython_env_in is not None:
        sPython_env = sPython_env_in
    else:
        sPython_env = 'base'

    if sQueue_in is not None:
        sQueue = sQueue_in
    else:
        sQueue = 'short'

    sWalltime = "{:0d}".format(iWalltime)
    sNode = "{:0d}".format(iNode)
    sThread = "{:0d}".format(nThread)

    os.chdir(sDirectory_job)

    ofs = open(sBasename_job, "w")  # write mode
    sLine = '#!/bin/bash' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH --account=esmd' + '\n'
    ofs.write(sLine)

    # sLine = '#SBATCH --begin=now+1minutes' + '\n'
    # ofs.write( sLine )

    sLine = '#SBATCH --cpus-per-task=1 ' + '\n'
    ofs.write(sLine)

    sLine = '#SBATCH --dependency=singleton ' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH --error=stderr_%j.err' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH --job-name=' + sJob_name + \
        '  # create a name for your job' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH --mail-type=ALL' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH --mail-user=' + sEmail + '\n'
    ofs.write(sLine)

    sLine = '#SBATCH --nodes=' + sNode + ' # node count' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH --ntasks=' + sThread + ' # total number of tasks' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH --output=stdout_%j.out' + '\n'
    ofs.write(sLine)

    sLine = '#SBATCH --partition=' + sQueue + '\n'  # can be improved here
    ofs.write(sLine)
    sLine = '#SBATCH --time=' + sWalltime + \
        ':00:00   # total run time limit (HH:MM:SS)' + '\n'
    ofs.write(sLine)

    sLine = 'module purge' + '\n'
    ofs.write(sLine)
    sLine = 'module load parafly/2013' + '\n'
    ofs.write(sLine)
    sLine = 'module load anaconda3/2019.03' + '\n'
    ofs.write(sLine)
    sLine = 'source /share/apps/anaconda3/2019.03/etc/profile.d/conda.sh' + '\n'
    ofs.write(sLine)
    sLine = 'unset PYTHONHOME' + '\n'
    ofs.write(sLine)
    sLine = 'conda activate ' + sPython_env + '\n'
    ofs.write(sLine)

    sLine = 'ParaFly -c ' + sBasename_parafly + ' -CPU ' + \
        sThread + ' -failed_cmds rerun.txt' + '\n'
    ofs.write(sLine)

    sLine = 'echo " Job " ' + '${SLURM_JOBID}' + ' is launched' + '\n'
    ofs.write(sLine)

    sLine = 'conda deactivate' + '\n'
    ofs.write(sLine)

    sLine = 'echo "Finished"' + '\n'
    ofs.write(sLine)
    ofs.close()

    return



import numpy as np
def calculate_hexagon_area(dLength_edge_in):
    """
    Calculate the area of a hexagon grid

    Args:
        dLength_edge_in (float): The length of hexagon edge

    Returns:
        float: The hexagon area
    """
    area = 1.5 * np.sqrt(3.0) * dLength_edge_in
    return area



import numpy as np
import matplotlib.pyplot as plt
from pyearth.visual.color.create_diverge_rgb_color_hex import create_diverge_rgb_color_hex
from pyearth.visual.create_line_style import create_line_style


def cdf_plot_multiple_data(aData,
                           sFilename_out,
                           iSize_x_in=None,
                           iSize_y_in=None,
                           iDPI_in=None,
                           dMin_x_in=None,
                           dMax_x_in=None,
                           dSpace_x_in=None,
                           sLabel_x_in=None,
                           sLabel_y_in=None,
                           sTitle_in=None,
                           aLabel_legend_in=None):
    """
    Draw a histogram for single dataset
    """
    nData = len(aData)

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12
    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9
    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if dMin_x_in is not None:
        dMin_x = dMin_x_in
    else:
        dMin_x = np.min(aData)

    if dMax_x_in is not None:
        dMax_x = dMax_x_in
    else:
        dMax_x = np.max(aData)

    if dSpace_x_in is not None:
        dSpace_x = dSpace_x_in
    else:
        # it may be calculated
        pass

    if sLabel_x_in is not None:
        sLabel_x = sLabel_x_in
    else:
        sLabel_x = ''

    if sLabel_y_in is not None:
        sLabel_y = sLabel_y_in
    else:
        sLabel_y = ''

    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    if aLabel_legend_in is not None:
        aLabel_legend = aLabel_legend_in
    else:
        aLabel_legend = np.full(nData, '', dtype=str)

    fig = plt.figure(dpi=iDPI)
    fig.set_figwidth(iSize_x)
    fig.set_figheight(iSize_y)

    left, width = 0.15, 0.7
    bottom, height = 0.1, 0.7
    spacing = 0.005
    rect_histogram = [left, bottom, width, height]

    ax_cdf = plt.axes(rect_histogram)
    ax_cdf.tick_params(direction='in', top=True, right=True)

    labels = []
    handles = []
    # setup color
    aColor = create_diverge_rgb_color_hex(nData)
    aLine_style = create_line_style(nData)
    for i in range(0, nData):
        aData_slice = aData[i]
        good_index = np.where((aData_slice >= dMin_x) &
                              (aData_slice <= dMax_x))
        aData_slice = aData_slice[good_index]
        count, bins_count = np.histogram(aData_slice, bins=100)
        pdf = count / sum(count)
        cdf = np.cumsum(pdf)
        ax_cdf.plot(bins_count[1:], cdf, color=aColor[i],
                    label=aLabel_legend[i], linestyle=aLine_style[i])
        labels.append(aLabel_legend[i])
        # handles.append( mpl_patches.Rectangle((0, 0), 1, 1, fc=aColor[i], ec="white", lw=0, alpha=0)  )

    ax_cdf.set_xlim(dMin_x, dMax_x)
    ax_cdf.axis('on')
    ax_cdf.grid(which='major', color='white', linestyle='-', axis='y')

    ax_cdf.set_xlabel(sLabel_x, fontsize=15)
    ax_cdf.set_ylabel(sLabel_y, fontsize=15)

    ax_cdf.legend(labels, loc="lower right", fontsize=15, framealpha=0.7)

    ax_cdf.set_title(sTitle)

    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')




import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from pyearth.visual.scatter.scatter_lowess import scatter_lowess
from pyearth.visual.color.create_diverge_rgb_color_hex import create_diverge_rgb_color_hex


def scatter_plot_multiple_data(aData_x,
                               aData_y,
                               sFilename_out,
                               iFlag_miniplot_in=None,
                               iFlag_scientific_notation_x_in=None,
                               iFlag_scientific_notation_y_in=None,
                               iSize_x_in=None,
                               iSize_y_in=None,
                               iDPI_in=None,
                               iFlag_log_x_in=None,
                               iFlag_log_y_in=None,
                               dMin_x_in=None,
                               dMax_x_in=None,
                               dMin_y_in=None,
                               dMax_y_in=None,
                               dSpace_x_in=None,
                               dSpace_y_in=None,
                               sFormat_x_in=None,
                               sFormat_y_in=None,
                               sLabel_x_in=None,
                               sLabel_y_in=None,
                               aLabel_point_in=None,
                               aColor_in=None,
                               aMarker_in=None,
                               aLabel_legend_in=None,
                               aSize_in=None,
                               sTitle_in=None,
                               sFont_in=None):
    # number of dataset

    # aData_in = np.array(aData_y)
    try:
        from scipy.stats import gaussian_kde
        from scipy import stats
    except ImportError as e:
        raise ImportError(
            "The package 'scipy' is required for this function to run.") from e

    nData = len(aData_y)

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 12

    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_miniplot_in is not None:
        iFlag_miniplot = iFlag_miniplot_in
    else:
        iFlag_miniplot = 0

    if iFlag_scientific_notation_x_in is not None:
        iFlag_scientific_notation_x = iFlag_scientific_notation_x_in
    else:
        iFlag_scientific_notation_x = 0

    if iFlag_scientific_notation_y_in is not None:
        iFlag_scientific_notation_y = iFlag_scientific_notation_y_in
    else:
        iFlag_scientific_notation_y = 0

    if iFlag_log_x_in is not None:
        iFlag_log_x = iFlag_log_x_in
    else:
        iFlag_log_x = 0

    if iFlag_log_y_in is not None:
        iFlag_log_y = iFlag_log_y_in
    else:
        iFlag_log_y = 0

    if sLabel_x_in is not None:
        sLabel_X = sLabel_x_in
    else:
        sLabel_X = ''

    if sLabel_y_in is not None:
        sLabel_Y = sLabel_y_in
    else:
        sLabel_Y = ''

    if aLabel_legend_in is not None:
        aLabel_legend = aLabel_legend_in
    else:
        aLabel_legend = ''

    if sFont_in is not None:
        sFont = sFont_in
    else:
        sFont = "Times New Roman"

    plt.rcParams["font.family"] = sFont
    plt.rcParams["mathtext.fontset"] = 'dejavuserif'

    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    fig = plt.figure(dpi=iDPI)
    fig.set_figwidth(iSize_x)
    fig.set_figheight(iSize_y)

    left, width = 0.1, 0.75
    bottom, height = 0.1, 0.75
    spacing = 0.005
    dY_mini = 0.55
    dX_mini = 0.12
    width_mini = 0.28
    rect_scatter = [left, bottom, width, height]
    rect_scatter_mini = [dY_mini, dX_mini, width_mini, width_mini]
    rect_histx = [left, bottom + height + spacing, width, 0.15]
    rect_histy = [left + width + spacing, bottom, 0.15, height]

    # sns.regplot(x, y, lowess=True)
    # ax_scatter = sns.regplot(x=aData_x, y=aData_y, marker="+", lowess=True)
    ax_scatter_full = plt.axes(rect_scatter)
    if iFlag_miniplot == 1:
        ax_scatter_mini = plt.axes(rect_scatter_mini)
        ax_scatter_all = [ax_scatter_full, ax_scatter_mini]
    else:
        ax_scatter_all = [ax_scatter_full]

    for i in range(nData):
        du = np.array(aData_x[i])
        dv = np.array(aData_y[i])
        du = du.flatten()
        dv = dv.flatten()
        if (i == 0):
            dummyx = du
            dummyy = dv
        else:
            dummyx = np.concatenate([dummyx, du])
            dummyy = np.concatenate([dummyy, dv])

    x_min = np.min([dummyx, dummyy])
    x_max = np.max([dummyx, dummyy])
    y_min = np.min([dummyx, dummyy])
    y_max = np.max([dummyx, dummyy])

    if aColor_in is None:
        # need a better solution here
        if (nData >= 3):
            if (nData <= 10):
                aColor = create_diverge_rgb_color_hex(nData)
            else:
                # we will use both symbol and color
                aColor = create_diverge_rgb_color_hex(5)
                nMarker = np.ceil(nData/5)
                aMarker = ['o', '+', 'x']

                pass
        else:
            if nData == 2:
                aColor = ['red', 'blue']
            else:
                aColor = ['red']
    else:
        aColor = aColor_in

    if aMarker_in is None:
        pass
    else:
        aMarker = aMarker_in

    if aSize_in is None:
        a = mpl.rcParams['lines.markersize'] ** 2
        aSize = np.full(nData, a, dtype=float)
    else:
        aSize = aSize_in

    if dMin_x_in is not None:
        dMin_x = dMin_x_in
    else:
        dMin_x = x_min

    if dMax_x_in is not None:
        dMax_x = dMax_x_in
    else:
        dMax_x = x_max

    if dMin_y_in is not None:
        dMin_y = dMin_y_in
    else:
        dMin_y = y_min

    if dMax_y_in is not None:
        dMax_y = dMax_y_in
    else:
        dMax_y = y_max

    if dSpace_x_in is not None:
        dSpace_x = dSpace_x_in
    else:
        dSpace_x = (dMax_x - dMin_x) / 4.0

    if dSpace_y_in is not None:
        dSpace_y = dSpace_y_in
    else:
        dSpace_y = (dMax_y - dMin_y) / 4.0

    # for ax_scatter in ax_scatter_all:
    dRatio = (float(iSize_y)/iSize_x) / ((dMax_y-dMin_y) / (dMax_x-dMin_x))
    for iax in range(len(ax_scatter_all)):
        ax_scatter = ax_scatter_all[iax]
        ax_scatter.tick_params(direction='in', top=True, right=True)

        if iax == 0:
            aLegend_artist = []
            aLabel = []
        else:
            pass

        for i in range(nData):
            x = aData_x[i]
            y = aData_y[i]

            # treatment for nan data
            x = np.array(x)
            y = np.array(y)
            x = x.flatten()
            y = y.flatten()
            a = np.logical_and(~np.isnan(x), ~np.isnan(y))
            x = x[a]
            y = y[a]

            sc = ax_scatter.scatter(
                x, y, s=aSize[i], alpha=0.5, color=aColor[i], marker=aMarker[i])
            if iax == 0:
                aLegend_artist.append(sc)
                slope, intercept, r_value, p_value, std_err = stats.linregress(
                    x, y)
                sR = "slope:" + \
                    "{:.2f}".format(slope) + r"; $r^2$:" + \
                    "{:.2f}".format(r_value**2)
                aLabel.append(aLabel_legend[i] + ': ' + sR)

                if i == 0 and aLabel_point_in is not None:

                    for j in range(len(x)):
                        ax_scatter.annotate(
                            aLabel_point_in[j], (x[j], y[j]), fontsize=16)
                    pass
            # ax_scatter.set_facecolor('silver')

        if iax == 0:

            ax_scatter.axis('on')
            ax_scatter.grid(which='major', color='grey',
                            linestyle='--', axis='y')

            ax_scatter.tick_params(axis="x", labelsize=13)
            ax_scatter.tick_params(axis="y", labelsize=13)

            ax_scatter.set_xmargin(0.05)
            ax_scatter.set_ymargin(0.15)

            ax_scatter.set_xlabel(sLabel_X, fontsize=16)
            ax_scatter.set_ylabel(sLabel_Y, fontsize=16)
            ax_scatter.set_title(sTitle, loc='center', fontsize=20)
            # round to nearest years...

            if sFormat_x_in is not None:
                sFormat_x = sFormat_x_in
                ax_scatter.xaxis.set_major_formatter(
                    mpl.ticker.FormatStrFormatter(sFormat_x))

                # ax_scatter.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1e'))'%.1f'

            if sFormat_y_in is not None:
                sFormat_y = sFormat_y_in
                ax_scatter.yaxis.set_major_formatter(
                    mpl.ticker.FormatStrFormatter(sFormat_y))

                # ax_scatter.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))'%.1f'

            ax_scatter.tick_params(axis='y', pad=8)
            ax_scatter.set_xlim(dMin_x, dMax_x * 1.05)
            ax_scatter.set_ylim(dMin_x, dMax_x * 1.05)
        else:
            ax_scatter.set_xlim(dMin_x, dMax_x * 0.1)
            ax_scatter.set_ylim(dMin_x, dMax_x * 0.1)
            pass

        ax_scatter.xaxis.set_major_locator(
            mpl.ticker.MaxNLocator(prune='upper', nbins=5))

        if iFlag_log_x == 1:
            aLabel_x = []
            for i in np.arange(dMin_x, dMax_x + 1, 1):
                sTicklabel = r'$10^{{{}}}$'.format(int(i))
                aLabel_x.append(sTicklabel)
                pass

            ax_scatter.set_xticks(np.arange(dMin_x, dMax_x + 1, 1))
            ax_scatter.set_xticklabels(aLabel_x)
        else:
            if iFlag_scientific_notation_x == 1:
                formatter = mpl.ticker.ScalarFormatter(useMathText=True)
                formatter.set_scientific(True)
                # you might need to change here
                formatter.set_powerlimits((-1, 1))
                ax_scatter.xaxis.set_major_formatter(formatter)
            else:
                pass

        # ax_scatter.yaxis.set_major_locator(mpl.ticker.MultipleLocator(base = dSpace_y))
        ax_scatter.yaxis.set_major_locator(
            mpl.ticker.MaxNLocator(prune='upper', nbins=5))

        if iFlag_log_y == 1:
            aLabel_y = []
            for i in np.arange(dMin_y, dMax_y + 1, 1):
                sTicklabel = r'$10^{{{}}}$'.format(int(i))
                aLabel_y.append(sTicklabel)
                pass

            ax_scatter.set_yticks(np.arange(dMin_y, dMax_y + 1, 1))
            ax_scatter.set_yticklabels(aLabel_y)
            pass
        else:
            if iFlag_scientific_notation_y == 1:
                formatter = mpl.ticker.ScalarFormatter(useMathText=True)
                formatter.set_scientific(True)
                # you might need to change here
                formatter.set_powerlimits((-1, 1))
                ax_scatter.yaxis.set_major_formatter(formatter)

            pass

        ax_scatter.set_aspect(dRatio)  # this one set the y / x ratio

        line = mpl.lines.Line2D([0, 1], [0, 1], color='black', linestyle='dashed')
        transform = ax_scatter.transAxes
        line.set_transform(transform)
        ax_scatter.add_line(line)

        iFlag_lowess = 0
        if (iFlag_lowess == 1):

            y_sm, y_std, order = scatter_lowess(aData_x, aData_y, f=1./3.)
            ax_scatter.plot(x[order], y_sm[order], color='tomato')
            ax_scatter.fill_between(x[order],
                                    y_sm[order] - 1.96*y_std[order],
                                    y_sm[order] + 1.96*y_std[order],
                                    alpha=0.3)

            sLabel_legend_lowess2 = 'LOWESS uncertainty'

            # labels.append(sLabel_legend_lowess2)

        if iax == 0:

            ax_scatter.legend(aLegend_artist, aLabel,
                              loc="upper left", fontsize=16)

        ax_scatter.tick_params(which='both',  # Options for both major and minor ticks
                               top='off',  # turn off top ticks
                               left='off',  # turn off left ticks
                               right='off',  # turn off right ticks
                               bottom='off')  # turn off bottom ticks

        if iax == 0:
            for i in range(nData):
                x = aData_x[i].flatten()
                # treatment for nan data
                a = np.where(~np.isnan(x))
                x = x[a]

                try:
                    if np.max(x) > np.min(x):
                        density = gaussian_kde(x)
                    else:
                        return
                except ValueError:
                    pass

                xx = np.linspace(dMin_x, dMax_x, 1000)
                yy = density(xx)

                # y margin
                y = aData_y[i].flatten()
                a = np.where(~np.isnan(y))
                y = y[a]

                density = gaussian_kde(y)
                xx = np.linspace(dMin_y, dMax_y, 1000)
                yy = density(xx)
                xx, yy = yy, xx

        if iax == 0:

            if iFlag_miniplot == 1:
                # horizontal
                line = mpl.lines.Line2D(
                    [0.0, 0.1], [0.1, 0.1], color='black', linestyle='dotted')
                transform = ax_scatter.transAxes
                line.set_transform(transform)
                ax_scatter.add_line(line)

                # vertical
                line = mpl.lines.Line2D(
                    [0.1, 0.1], [0.1, 0.0], color='black', linestyle='dotted')
                transform = ax_scatter.transAxes
                line.set_transform(transform)
                ax_scatter.add_line(line)

                line = mpl.lines.Line2D([0.1, (dY_mini-left)/width], [
                                     0.1, (dX_mini-bottom + width_mini)/height], color='black', linestyle='dotted')
                transform = ax_scatter.transAxes
                line.set_transform(transform)
                ax_scatter.add_line(line)

                line = mpl.lines.Line2D([0.1,  (dY_mini-left)/width], [0.0,
                                     (dX_mini-bottom)/height], color='black', linestyle='dotted')
                transform = ax_scatter.transAxes
                line.set_transform(transform)
                ax_scatter.add_line(line)

    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')
    print('finished plotting')



import os
import numpy as np

def text_reader_string(sFilename_in,
                       ncolumn_in=None,
                       nrow_in=None,
                       cDelimiter_in=None,
                       iFlag_remove_quota=None,
                       iSkipline_in=None):
    """
    Read a text based file

    Args:
        sFilename_in (string): The text filename
        ncolumn_in (int, optional): The number of column. Defaults to None.
        nrow_in (int, optional): The number of row. Defaults to None.
        cDelimiter_in (char, optional): The char to seperate a line. Defaults to None.
        iFlag_remove_quota (int, optional): The flag to remove the quota in line string. Defaults to None.
        iSkipline_in (int, optional): The line count to skip. Defaults to None.

    Returns:
        numpy.arry: The data inside the text file.
    """
    if os.path.exists(sFilename_in):
        pass
    else:
        print('The file does not exist!')
        return

    aData_out = -1

    # check the parameter
    if ncolumn_in is not None:
        iFlag_column = 1
        ncolumn_out = ncolumn_in
    else:
        iFlag_column = 0

    if nrow_in is not None:
        # there is nothing
        nrow_out = nrow_in
        # print(' ')
    else:
        # get the total line number
        ifs = open(sFilename_in, "r", errors='ignore')
        nrow_out = len(ifs.readlines())
        ifs.close()
    sLine = ' '
    ifs = open(sFilename_in, "r", errors='ignore')
    if iSkipline_in is not None:
        nrow_out = nrow_out - iSkipline_in
        for i in range(iSkipline_in):
            ifs.readline()
    else:
        pass
    if iFlag_remove_quota is not None:
        iFlag_iFlag_remove_quota = 1
    else:
        iFlag_iFlag_remove_quota = 0
    # get delimiter
    if cDelimiter_in is not None:
        iFlag_delimiter = 1
    else:
        iFlag_delimiter = 0
        # cDelimiter_in = ' '
        pass

    if iFlag_delimiter == 1:
        if iFlag_column == 1:
            if ncolumn_out < 1:
                print('The file has no data!')
                pass
            aData_out = np.full((nrow_out, ncolumn_out), '  ', dtype=object)
            for iRow in range(nrow_out):
                sLine = (ifs.readline()).rstrip()
                if iFlag_iFlag_remove_quota == 1:
                    sLine.replace('"', '')
                else:
                    pass
                aDummy = sLine.split(cDelimiter_in)
                iDummy = len(aDummy)
                if iDummy == ncolumn_out:
                    aData_out[iRow] = aDummy
                else:
                    # something is missing
                    aDummy2 = np.full(ncolumn_out, '  ', dtype=object)
                    aDummy2[0: iDummy] = aDummy
                    aData_out[iRow] = aDummy2
                    pass
        else:
            sLine = (ifs.readline()).rstrip()
            dummy = sLine.split(cDelimiter_in)
            ncolumn_out = len(dummy)
            # check ncolumn_in count
            if ncolumn_out < 1:
                print('The file has no data!')
                pass
            aData_out = np.full((nrow_out, ncolumn_out), '  ', dtype=object)
            aData_out[0] = dummy
            for iRow in range(1, nrow_out):
                sLine = (ifs.readline()).rstrip()
                if iFlag_iFlag_remove_quota == 1:
                    sLine.replace('"', '')
                else:
                    pass
                aDummy = sLine.split(cDelimiter_in)
                iDummy = len(aDummy)
                if iDummy == ncolumn_out:
                    aData_out[iRow] = aDummy
                else:
                    # something is missing
                    aDummy2 = np.full(ncolumn_out, '  ', dtype=object)
                    aDummy2[0: iDummy] = aDummy
                    aData_out[iRow] = aDummy2
                    pass

    else:
        if iFlag_column == 1:
            # check ncolumn_in count
            if ncolumn_out < 1:
                return aData_out
            aData_out = np.full((nrow_out, ncolumn_out), '  ', dtype=object)
            for iRow in range(nrow_out):
                dummy1 = ifs.readline()
                # print('New line is: ' + dummy1)
                dummy2 = dummy1.rstrip()
                if iFlag_iFlag_remove_quota == 1:
                    dummy2 = dummy2.replace('"', '')
                else:
                    pass
                dummy3 = dummy2.split()
                # print(dummy3)
                aData_out[iRow] = dummy3
        else:
            sLine = (ifs.readline()).rstrip()
            dummy = sLine.split()
            ncolumn_out = len(dummy)
            if ncolumn_out < 1:
                return aData_out
            aData_out = np.full((nrow_out, ncolumn_out), '  ', dtype=object)
            if iFlag_iFlag_remove_quota == 1:
                sLine = sLine.replace('"', '')
            else:
                pass
            
            aData_out[0] = dummy
            for iRow in range(1, nrow_out):
                dummy1 = ifs.readline()
                # print('New line is: ' + dummy1)
                dummy2 = dummy1.rstrip()
                if iFlag_iFlag_remove_quota == 1:
                    dummy2 = dummy2.replace('"', '')
                else:
                    pass
                dummy3 = dummy2.split()
                # print(dummy3)
                aData_out[iRow] = dummy3
    # print(aData_out)
    ifs.close()

    return aData_out



# plot a map of study area which has google earth image, slope, and dem
import numpy as np
from osgeo import osr, gdal, ogr

import matplotlib as mpl
import matplotlib.pyplot as plt
import cartopy as cpl

from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter

from pyearth.gis.gdal.read.raster.gdal_read_geotiff_file import gdal_read_geotiff_file
from pyearth.gis.spatialref.reproject_coodinates import reproject_coordinates, reproject_coordinates_batch
from pyearth.gis.location.Google_MetersPerPixel import Google_MetersPerPixel
from pyearth.gis.location.get_geometry_coordinates import get_geometry_coordinates


def plot_study_area(sFilename_dem_in,
                    sFilename_out,
                    sFilename_polygon_in=None,
                    sFilename_slope_in=None,
                    sFilename_point_in=None,
                    sFilename_polyline_in=None):

    try:
        import requests
    except ImportError as e:
        raise ImportError(
            "The package 'requests' is required for this function to run.") from e

    # set up figure size and dpi
    fig = plt.figure(dpi=300)
    fig.set_figwidth(4)
    fig.set_figheight(4)
    # ==============================================================================
    # plot dem first
    # ==============================================================================

    dummy = gdal_read_geotiff_file(sFilename_dem_in)
    aImage_in = dummy['dataOut']
    dResolution_x = dummy['pixelWidth']
    dResolution_y = dummy['pixelHeight']
    dOriginX = dummy['originX']
    dOriginY = dummy['originY']
    nrow = dummy['nrow']
    ncolumn = dummy['ncolumn']
    missing_value = dummy['missingValue']
    pProjection = dummy['projection']
    pSpatialRef_source = dummy['spatialReference']

    # set up dem projection
    dem_proj = cpl.crs.AlbersEqualAre(central_longitude=pSpatialRef_source.GetProjPar('longitude_of_center'),
                                   central_latitude=pSpatialRef_source.GetProjPar(
                                       'latitude_of_center'),
                                   standard_parallels=(pSpatialRef_source.GetProjPar('standard_parallel_1'),
                                                       pSpatialRef_source.GetProjPar('standard_parallel_2')))

    # set up dem plot
    ax_dem = fig.add_axes([0.3, 0.15, 0.6, 0.55], projection=dem_proj)
    ax_dem.set_xmargin(0.05)
    ax_dem.set_ymargin(0.10)

    dLon_min = dOriginX
    dLon_max = dOriginX + ncolumn * dResolution_x
    dLat_max = dOriginY
    dLat_min = dOriginY + nrow * dResolution_y

    pSpatialRef_target = osr.SpatialReference()
    pSpatialRef_target.ImportFromEPSG(4326)
    aLon = list()
    aLat = list()
    aLon.append(dLon_min)
    aLon.append(dLon_max)
    aLat.append(dLat_min)
    aLat.append(dLat_max)
    aLon, aLat = reproject_coordinates_batch(
        aLon, aLat, pSpatialRef_source, pSpatialRef_target)
    dLongitude_center = np.mean(aLon)
    dLatitude_center = np.mean(aLat)
    aImage_extent = [dLon_min - dResolution_x, 
                     dLon_max + dResolution_x, 
                     dLat_min + dResolution_y,  
                     dLat_max - dResolution_y]

    aImage_in[np.where(aImage_in == missing_value)] = np.nan
    #
    demplot = ax_dem.imshow(aImage_in, origin='upper',
                            extent=aImage_extent, cmap=mpl.cm.terrain, transform=dem_proj)
    # change all spines
    for axis in ['top', 'bottom', 'left', 'right']:
        ax_dem.spines[axis].set_linewidth(0.5)
        # increase tick width
    ax_dem.tick_params(width=0.5)

    # plot flowline if available
    pDriver = ogr.GetDriverByName('GeoJSON')
    if sFilename_polyline_in is None:
        pass
    else:
        pDataset = pDriver.Open(sFilename_polyline_in, gdal.GA_ReadOnly)
        pLayer = pDataset.GetLayer(0)
        for pFeature in pLayer:
            pGeometry_in = pFeature.GetGeometryRef()
            sGeometry_type = pGeometry_in.GetGeometryName()
            if sGeometry_type == 'LINESTRING':
                # dummy0 = loads( pGeometry_in.ExportToWkt() )
                # aCoords_gcs = dummy0.coords
                aCoords_gcs = get_geometry_coordinates(pGeometry_in)
                aCoords_gcs = np.array(aCoords_gcs)
                aCoords_gcs = aCoords_gcs[:, 0:2]
                nvertex = len(aCoords_gcs)

                codes = np.full(nvertex, mpl.path.Path.LINETO, dtype=int)
                codes[0] = mpl.path.Path.MOVETO
                path = mpl.path.Path(aCoords_gcs, codes)
                x, y = zip(*path.vertices)
                line, = ax_dem.plot(
                    x, y, color='b', transform=cpl.crs.PlateCarree())
                lID = lID + 1

    # plot point if available
    if sFilename_point_in is None:
        pass
    else:
        pDataset = pDriver.Open(sFilename_point_in, gdal.GA_ReadOnly)
        pLayer = pDataset.GetLayer(0)
        for pFeature in pLayer:
            pGeometry_in = pFeature.GetGeometryRef()
            sGeometry_type = pGeometry_in.GetGeometryName()
            if sGeometry_type == 'POINT':
                aCoords_gcs = get_geometry_coordinates(pGeometry_in)
                aCoords_gcs = np.array(aCoords_gcs)
                aCoords_gcs = aCoords_gcs[:, 0:2]
                nvertex = len(aCoords_gcs)

                codes = np.full(nvertex, mpl.path.Path.LINETO, dtype=int)
                codes[0] = mpl.path.Path.MOVETO
                path = mpl.path.Path(aCoords_gcs, codes)
                x, y = zip(*path.vertices)
                line, = ax_dem.plot(
                    x, y, color='b', transform=cpl.crs.PlateCarree())
                lID = lID + 1

    # draw gridlines
    gl = ax_dem.gridlines(crs=cpl.crs.PlateCarree(), draw_labels=True,
                          linewidth=1, color='gray', alpha=0.3, linestyle='--')

    gl.left_labels = False
    gl.top_labels = False
    XTEXT_SIZE = 8
    YTEXT_SIZE = 8
    # to facilitate text rotation at bottom edge, ...
    # text justification: 'ha':'right' is used to avoid clashing with map'sboundary
    # default of 'ha' is center, often causes trouble when text rotation isnot zero
    gl.xlabel_style = {'size': XTEXT_SIZE,
                       'color': 'k', 'rotation': 0, 'ha': 'right'}
    gl.ylabel_style = {'size': YTEXT_SIZE, 'color': 'k',
                       'rotation': 90, 'weight': 'normal'}

    # draw colorbar
    ax_cb = fig.add_axes([0.2, 0.2, 0.02, 0.5])
    cb = plt.colorbar(demplot, cax=ax_cb, extend='both')
    cb.ax.get_yaxis().set_ticks_position('left')
    cb.ax.get_yaxis().labelpad = 10
    cb.ax.set_ylabel('Unit: meter', rotation=270)
    cb.ax.tick_params(labelsize=6)

    # ==============================================
    # google earth
    ge_proj = cpl.crs.Mercator()  # central_longitude=dLongitude_center)

    pSpatialRef_source = osr.SpatialReference()
    pSpatialRef_source.ImportFromEPSG(4326)
    pSpatialRef_target = osr.SpatialReference()

    iZoom = 7
    pSpatialRef_target.ImportFromWkt(ge_proj.to_wkt())
    uv_xcenter, uv_ycenter = reproject_coordinates(
        dLongitude_center, dLatitude_center, pSpatialRef_source, pSpatialRef_target)
    xsize_ge = 600
    ysize_ge = 600
    scale = Google_MetersPerPixel(iZoom)
    xrange = (uv_xcenter - (xsize_ge/2.0*scale),
              uv_xcenter + (xsize_ge/20*scale))
    yrange = (uv_ycenter - (ysize_ge/2.0*scale),
              uv_ycenter + (ysize_ge/20*scale))
    aImage_extent = [xrange[0], xrange[1], yrange[0], yrange[1]]
    ax_ge = fig.add_axes([0.1, 0.75, 0.2, 0.2], projection=ge_proj)

    sLongitude_center = "{:0f}".format(dLongitude_center)
    sLatitude_center = "{:0f}".format(dLatitude_center)
    sZoom = "{:0d}".format(iZoom)

    sResolution = "{:0d}".format(xsize_ge) + 'x' + "{:0d}".format(ysize_ge)
    sMap_type = "hybrid"
    #obtain GCP APIs from your system
    sGoogleMapAPI = ''
    sGoogleMap = "http://maps.googleapis.com/maps/api/staticmap?" + \
        "center=" + sLatitude_center + ',' + sLongitude_center + \
        "&zoom=" + sZoom + "&size=" + sResolution + \
        "&maptype="+sMap_type+"&sensor=false&format=png32" + \
        "key=" + sGoogleMapAPI
    r = requests.get(sGoogleMap)
    # wb mode is stand for write binary mode
    f = open('google_map.png', 'wb')
    # r.content gives content,
    # in this case gives image
    f.write(r.content)
    # close method of file object
    # save and close the file
    f.close()

    img = mpl.image.imread('google_map.png')
    ax_ge.imshow(img, extent=aImage_extent,       transform=ge_proj)
    gl_ge = ax_ge.gridlines(crs=cpl.crs.PlateCarree(), draw_labels=True,
                            linewidth=1, color='gray', alpha=0.3, linestyle='--')
    # gl_ge.xlocator = mpl.ticker.FixedLocator(aLon_range)
    # gl_ge.ylocator = mpl.ticker.FixedLocator(aLat_range)
    gl_ge.xformatter = LongitudeFormatter()
    gl_ge.yformatter = LatitudeFormatter()
    gl_ge.right_labels = False
    gl_ge.bottom_labels = False

    # declare text size
    XTEXT_SIZE = 2.5
    YTEXT_SIZE = 2.5
    gl_ge.xlabel_style = {'size': XTEXT_SIZE,
                          'color': 'k', 'rotation': 0, 'ha': 'right'}
    gl_ge.ylabel_style = {'size': YTEXT_SIZE,
                          'color': 'k', 'rotation': 90, 'weight': 'normal'}
    # add boundary
    if sFilename_polygon_in is None:
        pass
    else:
        pDataset = pDriver.Open(sFilename_polygon_in, gdal.GA_ReadOnly)
        pLayer = pDataset.GetLayer(0)
        for pFeature in pLayer:
            pGeometry_in = pFeature.GetGeometryRef()
            sGeometry_type = pGeometry_in.GetGeometryName()
            if sGeometry_type == 'POLYGON':
                aCoords_gcs = get_geometry_coordinates(pGeometry_in)
                aCoords_gcs = np.array(aCoords_gcs)
                polygon = mpl.patches.Polygon(aCoords_gcs[:, 0:2], closed=True, linewidth=0.25,
                                           alpha=0.8, edgecolor='b',
                                           transform=cpl.crs.PlateCarree())
                ax_ge.add_patch(polygon)

    # histogram
    ax_histo = fig.add_axes([0.4, 0.8, 0.4, 0.15])
    if sFilename_slope_in is None:
        pass
    else:
        dummy = gdal_read_geotiff_file(sFilename_slope_in)
        aSlope = dummy[0]
        missing_value = dummy[7]
        aSlope = aSlope[np.where(aSlope != missing_value)]
        dMax_x = 60
        dMin_x = 0
        dSpace_x = 4
        ax_histo.hist(aSlope,  int((dMax_x-dMin_x) / dSpace_x),
                      color="skyblue", ec="skyblue")
        sLabel_x = 'Slope (degree)'
        sLabel_y = 'Frequency'
        ax_histo.set_xlabel(sLabel_x, fontsize=4)
        ax_histo.set_ylabel(sLabel_y, fontsize=4)
        ax_histo.set_xlim(dMin_x, dMax_x)
        formatter = mpl.ticker.ScalarFormatter(useMathText=True)
        formatter.set_scientific(True)
        ax_histo.yaxis.set_major_formatter(formatter)
        ax_histo.tick_params(axis='x', labelsize=5)
        ax_histo.tick_params(axis='y', labelsize=5)
        ax_histo.yaxis.offsetText.set_fontsize(5)

    plt.savefig(sFilename_out, bbox_inches='tight')

    return



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from pyearth.system.define_global_variables import *
from pyearth.visual.color.create_diverge_rgb_color_hex import create_diverge_rgb_color_hex
from pyearth.visual.color.choose_n_color import polylinear_gradient, rand_hex_color


def histogram_plot(aData_all,
                   sFilename_out,
                   iSize_x_in=None,
                   iSize_y_in=None,
                   ncolumn_in=None,
                   iFlag_scientific_notation_in=None,
                   iFlag_log_in=None,
                   aColor_in=None,
                   iDPI_in=None,
                   dMin_x_in=None,
                   dMax_x_in=None,
                   dSpace_x_in=None,
                   sLabel_x_in=None,
                   sLabel_y_in=None,
                   sFormat_x_in=None,
                   aLocation_legend_in=None,
                   sLocation_legend_in=None,
                   sTitle_in=None,
                   aLabel_legend_in=None):
    """
    Draw a histogram for single/multiple dataset
    """

    try:
        import scipy
    except ImportError as e:
        raise ImportError(
            "The package 'scipy' is required for this function to run.") from e

    aData_all = np.array(aData_all)
    pShape = aData_all.shape

    nData = pShape[0]

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12
    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9
    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_scientific_notation_in is not None:
        iFlag_scientific_notation = iFlag_scientific_notation_in
    else:
        iFlag_scientific_notation = 0

    if iFlag_log_in is not None:
        iFlag_log = iFlag_log_in
    else:
        iFlag_log = 0

    if sFormat_x_in is not None:
        iFlag_format_x = 1
        sFormat_x = sFormat_x_in
    else:
        iFlag_format_x = 0

    if dMin_x_in is not None:
        dMin_x = dMin_x_in
    else:
        dMin_x = np.min(aData_all)

    if dMax_x_in is not None:
        dMax_x = dMax_x_in
    else:
        dMax_x = np.max(aData_all)

    if dSpace_x_in is not None:
        iFlag_space_x = 0
        dSpace_x = dSpace_x_in
    else:
        iFlag_space_x = 1
        dSpace_x = (dMax_x - dMin_x) / 10
        pass

    if sLocation_legend_in is not None:
        sLocation_legend = sLocation_legend_in
    else:
        sLocation_legend = "upper right"

    if aLocation_legend_in is not None:
        aLocation_legend = aLocation_legend_in
    else:
        aLocation_legend = (1.0, 1.0)

    if sLabel_x_in is not None:
        sLabel_x = sLabel_x_in
    else:
        sLabel_x = ''

    if sLabel_y_in is not None:
        sLabel_y = sLabel_y_in
    else:
        sLabel_y = ''

    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    if aLabel_legend_in is not None:
        aLabel_legend = aLabel_legend_in
    else:
        aLabel_legend = np.full(nData, '')

    if aColor_in is not None:
        aColor = aColor_in
    else:
        if (nData >= 3):
            if (nData <= 12):
                aColor = create_diverge_rgb_color_hex(nData)
            else:
                a = rand_hex_color(num=2)
                b = polylinear_gradient(a, nData)
                aColor = b['hex']
                pass
        else:
            if nData == 2:
                aColor = ['red', 'blue']
            else:
                aColor = ['lightblue']

    fig = plt.figure(dpi=iDPI)
    fig.set_figwidth(iSize_x)
    fig.set_figheight(iSize_y)

    left, width = 0.15, 0.7
    bottom, height = 0.1, 0.85
    spacing = 0.005
    rect_histogram = [left, bottom, width, height]

    ax_histo = plt.axes(rect_histogram)
    ax_histo.tick_params(direction='in', top=True, right=True)

    aLegend_artist = []
    aLabel = []

    iFlag_method = 2
    if iFlag_method == 1:  # this method possible won't work for multi-case
        for i in np.arange(1, nData+1):
            aData = aData_all[i-1]
            bad_index = np.where(aData < dMin_x_in)
            aData[bad_index] = dMin_x_in
            bad_index = np.where(aData > dMax_x_in)
            aData[bad_index] = dMax_x_in

            if iFlag_log == 1:
                if dSpace_x >= 1:
                    dSpace_x = int(dSpace_x)
                else:
                    pass

            if i == 0:
                N, bins, hisp = ax_histo.hist(aData, int((dMax_x-dMin_x)/dSpace_x),
                                              color=aColor[i-1],  label=aLabel_legend[i-1])
            else:
                N, bins, hisp = ax_histo.hist(aData, int((dMax_x-dMin_x)/dSpace_x),
                                              color=aColor[i-1],  label=aLabel_legend[i-1], alpha=.5)

            aLegend_artist.append(hisp)
            aLabel.append(aLabel_legend[i-1])
    else:

        bad_index = np.where(aData_all < dMin_x_in)
        aData_all[bad_index] = dMin_x_in
        bad_index = np.where(aData_all > dMax_x_in)
        aData_all[bad_index] = dMax_x_in

        if iFlag_log == 1:
            if dSpace_x >= 1:
                dSpace_x = int(dSpace_x)
            else:
                pass

        aData_all_t = np.transpose(aData_all)
        N, bins, hisp = ax_histo.hist(aData_all_t, density=True,
                                      color=aColor, label=aLabel_legend)

        # add density?
        for i in np.arange(1, nData+1):
            aData = aData_all[i-1]
            density = scipy.stats.gaussian_kde(aData)
            good_index = np.where((aData >= dMin_x) & (aData <= dMax_x))
            aData = aData[good_index]
            xx = np.linspace(dMin_x, dMax_x, 1000)
            yy = density(xx)
            ax_histo.plot(xx, yy, color=aColor[i-1], linestyle='--')

    ax_histo.set_xlabel(sLabel_x, fontsize=13)
    ax_histo.set_ylabel(sLabel_y, fontsize=13)
    ax_histo.set_xlim(dMin_x, dMax_x)
    ax_histo.axis('on')

    if iFlag_log == 1:
        aLabel_x = list()
        xtickslocs = ax_histo.get_xticks().tolist()
        for i in np.arange(0, len(xtickslocs), 1):
            ii = xtickslocs[i]
            iii = sFormat_x.format(ii)
            sTicklabel = r'$10^{{{}}}$'.format(iii)
            aLabel_x.append(sTicklabel)
            pass

        ax_histo.set_xticks(xtickslocs)
        ax_histo.set_xticklabels(aLabel_x)
        pass
    else:
        if iFlag_scientific_notation == 1:
            formatter = mpl.ticker.ScalarFormatter(useMathText=True)
            formatter.set_scientific(True)
            ax_histo.yaxis.set_major_formatter(formatter)
            pass
        else:
            if (iFlag_format_x == 1):
                sFormat_x_dummy = sFormat_x.replace("{", "{x")
                ax_histo.xaxis.set_major_formatter(
                    mpl.ticker.StrMethodFormatter(sFormat_x_dummy))
                pass

            pass
        pass

    iFlag_grid = 0  # reserved option
    if iFlag_grid == 1:
        ax_histo.grid(which='major', color='white', linestyle='-', axis='y')

    if iFlag_method == 1:
        ax_histo.legend(aLegend_artist, aLabel, bbox_to_anchor=aLocation_legend,
                        loc=sLocation_legend, fontsize=12)
    else:
        ax_histo.legend(bbox_to_anchor=aLocation_legend,
                        loc=sLocation_legend, fontsize=12)

    ax_histo.set_title(sTitle)
    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')
    plt.clf()



import os
from osgeo import gdal

def get_gdal_driver_by_format(file_format):
    """
    Get the GDAL driver based on the provided raster file format.

    Args:
        file_format (str): The raster file format (e.g., 'GTiff', 'JP2', 'HFA').

    Returns:
        gdal.Driver: The GDAL driver object corresponding to the file format.
                      Returns None if the driver is not found.
    """
    # Get the list of GDAL drivers
    drivers = gdal.GetDriverCount()
    
    # Loop through the drivers and find the matching one
    for i in range(drivers):
        driver = gdal.GetDriver(i)
        if driver.GetMetadataItem(gdal.DCAP_RASTER) == 'YES':
            if file_format.upper() in [x.upper() for x in driver.GetMetadataItem(gdal.DMD_EXTENSION).split()]:
                return driver
    
    return None


import os
import numpy as np
from osgeo import  osr, gdal, ogr
from matplotlib.colors import ListedColormap
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.path as mpath
import matplotlib.cm as cm
from matplotlib.collections import PathCollection
from matplotlib.path import Path
from matplotlib.collections import PatchCollection
from matplotlib.collections import LineCollection
from matplotlib.patches import Polygon
import cartopy as cpl
import cartopy.crs as ccrs
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
from cartopy.io.img_tiles import OSM
from pyearth.toolbox.math.stat.remap import remap
from pyearth.gis.location.get_geometry_coordinates import get_geometry_coordinates
from pyearth.visual.formatter import OOMFormatter

#get the current year for openstreetmap copy right label
import datetime
iYear_current = datetime.datetime.now().year
#convert to string
sYear = str(iYear_current)

def map_multiple_vector_data(aFiletype_in,
                             aFilename_in,
                             iFlag_colorbar_in = None,
                             iFlag_title_in = None,
                             aFlag_thickness_in = None,
                             aFlag_color_in = None,
                             aFlag_discrete_in = None,
                             aFlag_fill_in = None,
                             aVariable_in = None,
                             sFilename_output_in=None,
                             iFlag_scientific_notation_colorbar_in = None,
                             iFlag_openstreetmap_in = None,
                             iFont_size_in=None,
                             iFlag_openstreetmap_level_in = None,
                             sColormap_in = None,
                             sTitle_in = None,
                             iDPI_in = None,
                             aMissing_value_in = None,
                             aData_max_in = None,
                             aData_min_in = None,
                             sExtend_in =None,
                             sFont_in = None,
                             sUnit_in=None,
                             aLegend_in = None,
                             aExtent_in = None,
                             pProjection_map_in=None):
    """
    plot vector data on a map
    currently only support geojson and shapefile
    by default, the program will plot all the polygons in the file
    in the furture, the program will support to plot only a subset of polygons

    Because of the overlay effect, it is ideal to plot them in the following order: polygon->polyling->point

    Args:
        iFiletype_in (_type_): _description_
        sFilename_in (_type_): _description_
        sFilename_output_in (_type_): _description_
        iFlag_scientific_notation_colorbar_in (_type_, optional): _description_. Defaults to None.
        sColormap_in (_type_, optional): _description_. Defaults to None.
        sTitle_in (_type_, optional): _description_. Defaults to None.
        iDPI_in (_type_, optional): _description_. Defaults to None.
        dMissing_value_in (_type_, optional): _description_. Defaults to None.
        dData_max_in (_type_, optional): _description_. Defaults to None.
        dData_min_in (_type_, optional): _description_. Defaults to None.
        sExtend_in (_type_, optional): _description_. Defaults to None.
        sUnit_in (_type_, optional): _description_. Defaults to None.
        aLegend_in (_type_, optional): _description_. Defaults to None.
    """

    #check vector type first
    if aFiletype_in is None:
        print('Error: please specify the vector type')
        return
    else:
        #point: 1, polyline: 2, polygon: 3      
        arr = np.array(aFiletype_in)      
        # Check if the array is descending or has the same values
        is_descending = np.all(np.diff(arr) <= 0)        
        if is_descending == True:
            pass
        else:
            print('Error: the vector type is not correct')
            return

    pDriver = ogr.GetDriverByName('GeoJSON')


    nFile = len(aFilename_in)
    if aFlag_thickness_in is None:
        aFlag_thickness= np.zeros(nFile, dtype=np.int16)
    else:
        aFlag_thickness = aFlag_thickness_in

    if aFlag_color_in is None:
        aFlag_color= np.zeros(nFile, dtype=np.int16)
    else:
        aFlag_color = aFlag_color_in
    
    if aFlag_discrete_in is None:
        aFlag_discrete= np.zeros(nFile, dtype=np.int16)
    else:
        aFlag_discrete = aFlag_discrete_in

    if aFlag_fill_in is None:
        aFlag_fill= np.zeros(nFile, dtype=np.int16)
    else:
        aFlag_fill = aFlag_fill_in

    #get the extent first
    sFilename_in = aFilename_in[0]

    pDataset = ogr.Open(sFilename_in, gdal.GA_ReadOnly)
    pLayer = pDataset.GetLayer(0)

    if iFlag_colorbar_in is not None:
        iFlag_colorbar = iFlag_colorbar_in
    else:
        iFlag_colorbar = 0
    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_title_in is not None:
        iFlag_title = iFlag_title_in
        if iFlag_title == 1:
            if sTitle_in is not None:
                sTitle = sTitle_in
            else:
                sTitle =  ''     

        else:
            iFlag_title=0

    else:
        iFlag_title = 0
        sTitle =  ''       
            
    
    if aMissing_value_in is not None:
        aMissing_value = aMissing_value_in
    else:
        aMissing_value =  np.full(nFile, -9999)
    if aData_min_in is not None:        
        aData_min = aData_min_in
        aFlag_data_min = np.full(nFile, 1)
    else:
        aData_min=np.full(nFile, -9999)
        aFlag_data_min = np.zeros(nFile, dtype=np.int16)    

    if aData_max_in is not None:
        aData_max = aData_max_in
        aFlag_data_max = np.full(nFile, 1)        
    else:
        aData_max = np.full(nFile, -9999)
        aFlag_data_max = np.zeros(nFile, dtype=np.int16)    
        pass    

    if iFlag_scientific_notation_colorbar_in is not None:
        iFlag_scientific_notation_colorbar = iFlag_scientific_notation_colorbar_in
    else:
        iFlag_scientific_notation_colorbar = 0

    if sColormap_in is not None:
        sColormap = sColormap_in
    else:
        sColormap =  'rainbow'   
    
    if iFont_size_in is not None:
        iFont_size = iFont_size_in
    else:
        iFont_size = 12

    if sExtend_in is not None:
        sExtend = sExtend_in
    else:
        sExtend =  'max'

    if sUnit_in is not None:
        sUnit = sUnit_in
    else:
        sUnit =  ''

    if sFont_in is not None:
        sFont = sFont_in
    else:
        sFont = "Times New Roman"

    plt.rcParams["font.family"] = sFont
    plt.rcParams["mathtext.fontset"] = 'dejavuserif'

    cmap = cm.get_cmap(sColormap)
    fig = plt.figure( dpi = iDPI  )
    iSize_x= 8
    iSize_y= 8
    fig.set_figwidth( iSize_x )
    fig.set_figheight( iSize_y )

    #we require that the first polygon file defines the extent
    pLayer = pDataset.GetLayer(0)
    pSrs = osr.SpatialReference()
    pSrs.ImportFromEPSG(4326)    # WGS84 lat/lon
    dLat_min = 90
    dLat_max = -90
    dLon_min = 180
    dLon_max = -180
    for pFeature in pLayer:
        pGeometry_in = pFeature.GetGeometryRef()
        sGeometry_type = pGeometry_in.GetGeometryName()
        if sGeometry_type =='MULTIPOLYGON':   
            for j in range(pGeometry_in.GetGeometryCount()):  
                pPolygon = pGeometry_in.GetGeometryRef(j)     
                aCoords_gcs = get_geometry_coordinates(pPolygon)
                dLon_max = np.max( [dLon_max, np.max(aCoords_gcs[:,0])] )
                dLon_min = np.min( [dLon_min, np.min(aCoords_gcs[:,0])] )
                dLat_max = np.max( [dLat_max, np.max(aCoords_gcs[:,1])] )
                dLat_min = np.min( [dLat_min, np.min(aCoords_gcs[:,1])] )
        else:
            if sGeometry_type =='POLYGON':            
                aCoords_gcs = get_geometry_coordinates(pGeometry_in)
                dLon_max = np.max( [dLon_max, np.max(aCoords_gcs[:,0])] )
                dLon_min = np.min( [dLon_min, np.min(aCoords_gcs[:,0])] )
                dLat_max = np.max( [dLat_max, np.max(aCoords_gcs[:,1])] )
                dLat_min = np.min( [dLat_min, np.min(aCoords_gcs[:,1])] )
            else:
                if sGeometry_type =='LINESTRING':
                    aCoords_gcs = get_geometry_coordinates(pGeometry_in)
                    dLon_max = np.max( [dLon_max, np.max(aCoords_gcs[:,0])] )
                    dLon_min = np.min( [dLon_min, np.min(aCoords_gcs[:,0])] )
                    dLat_max = np.max( [dLat_max, np.max(aCoords_gcs[:,1])] )
                    dLat_min = np.min( [dLat_min, np.min(aCoords_gcs[:,1])] )

    if pProjection_map_in is not None:
        pProjection_map = pProjection_map_in
    else:
        pProjection_map = ccrs.Orthographic(central_longitude =  0.50*(dLon_max+dLon_min),
                                            central_latitude = 0.50*(dLat_max+dLat_min),
                                            globe=None)

    ax = fig.add_axes([0.08, 0.1, 0.62, 0.7], projection= pProjection_map  ) #projection=ccrs.PlateCarree()

    # Create an OSM image tile source
    if aExtent_in is None:
        marginx  = (dLon_max - dLon_min) / 50
        marginy  = (dLat_max - dLat_min) / 50
        aExtent = [dLon_min - marginx , dLon_max + marginx , dLat_min -marginy , dLat_max + marginy]
    else:
        aExtent = aExtent_in
    
    ax.set_global()
    ax.set_extent( aExtent )
    if iFlag_openstreetmap_in is not None and iFlag_openstreetmap_in == 1:
        if iFlag_openstreetmap_level_in is not None:
            iFlag_openstreetmap_level = iFlag_openstreetmap_level_in
        else:
            iFlag_openstreetmap_level = 9
            pass

        osm_tiles = OSM()
        #Add the OSM image to the map
        ax.add_image(osm_tiles, iFlag_openstreetmap_level)   
        sLicense_info = "© OpenStreetMap contributors "+ sYear + "." + " Distributed under the Open Data Commons Open Database License (ODbL) v1.0."
        ax.text(0.5, 0.05, sLicense_info, transform=ax.transAxes, ha='center', va='center', fontsize=6, 
                color='gray', bbox=dict(facecolor='white', edgecolor='black', boxstyle='round,pad=0.3'))


    minx, miny, maxx, maxy = aExtent 
    #====================================
    #should we allow more than one scale for one variable?
    aValue_all = list()
    for i in range(nFile):
        aValue = list()
        sFilename = aFilename_in[i]
        iFlag_thickness = aFlag_thickness[i]
        iFlag_color = aFlag_color[i]
        sVariable = aVariable_in[i]
        pDataset = ogr.Open(sFilename, gdal.GA_ReadOnly)
        pLayer = pDataset.GetLayer(0)
        if len(sVariable) > 0:
            #select field value using sqlite        
            pKwargs = 'SELECT ' + sVariable + ' FROM ' + pLayer.GetName() 
            if iFlag_thickness ==1 : 
                pLayer_temp = pDataset.ExecuteSQL(pKwargs)
            else:
                if iFlag_color  == 1:
                    pLayer_temp = pDataset.ExecuteSQL(pKwargs)
            values = []
            # Iterate over the layer
            for feature in pLayer_temp:
                # Get the desired field value
                value = feature.GetField(sVariable)
                # Append the value to the list
                values.append(value)
            #conver aValue to numpy array
            aValue = np.array(values)
            aValue_all.append(aValue)      
        else:
            aValue_all.append(aValue)  

    iThickness_max = 2.5
    iThickness_min = 0.3

    #set min and max
    for i in range(nFile):
        iFlag_data_min = aFlag_data_min[i]
        iFlag_data_max = aFlag_data_max[i]
        dMissing_value = aMissing_value[i]
        aValue = np.array(aValue_all[i])
        aValue = aValue[aValue != dMissing_value]
        if len(aValue) == 0:
            aData_min[i] = 0
            aData_max[i] = 0
        else:
            if iFlag_data_min != 1:                          
                aData_min[i] = np.min(aValue)    
            if iFlag_data_max != 1:             
                aData_max[i] = np.max(aValue)        

    for i in range(nFile):
        sFilename = aFilename_in[i]
        iFlag_thickness = aFlag_thickness[i]
        iFlag_color = aFlag_color[i]
        iFlag_discrete = aFlag_discrete[i]
        iFlag_fill = aFlag_fill[i]
        dValue_min = aData_min[i]
        dValue_max = aData_max[i]
        dMissing_value = aMissing_value[i]
        sVariable = aVariable_in[i]
        aValue = np.array(aValue_all[i]) 
        pDataset = ogr.Open(sFilename, gdal.GA_ReadOnly)
        pLayer = pDataset.GetLayer(0)
        nValue = pLayer.GetFeatureCount()
        if iFlag_discrete == 1: #discrete
            aValue = np.array(aValue, dtype=int)
            #reorder it 
            aValue = np.unique(aValue)
            aValue = np.sort(aValue)
            nValue_discrete = len(aValue)
            aIndex = np.linspace(0,1,nValue_discrete)
            prng = np.random.RandomState(1234567890)
            prng.shuffle(aIndex)          
            colors = plt.cm.get_cmap(sColormap)(aIndex)
            aColorMap = ListedColormap(colors)
            pass
        else: #continuous
            aColorMap = plt.cm.get_cmap(sColormap)(np.linspace(0, 1, nValue))
        lID = 0
        aPoint=list()
        aPolyline=list()
        aThickness=list()
        aPolygon = list()
        aColor = list() 
        pLayer.SetSpatialFilterRect(minx, miny, maxx, maxy)
        
        for pFeature in pLayer:
            pGeometry_in = pFeature.GetGeometryRef()
            sGeometry_type = pGeometry_in.GetGeometryName()
            if iFlag_thickness ==1 :
                dValue = float(pFeature.GetField(sVariable))
                if dValue < dValue_min or dValue > dValue_max:
                    continue

                dValue = np.clip(dValue, dValue_min, dValue_max)

                iThickness = remap( dValue, dValue_min, dValue_max, iThickness_min, iThickness_max )
            else:
                iThickness = 0.25

            if iFlag_color == 1:          
                dValue = float(pFeature.GetField(sVariable))
                if iFlag_discrete ==1:          
                    if dValue < dValue_min or dValue > dValue_max:
                        continue           
                    iValue = int(dValue)                       
                    iColor_index = np.where(aValue == iValue)[0][0] 
                    sColor = aColorMap(iColor_index)
                else:    
                    if dValue < dValue_min or dValue > dValue_max:
                        continue
                    
                    iColor_index = int( (dValue - dValue_min) / (dValue_max - dValue_min) * 255 )
                    sColor = aColorMap[iColor_index]
            else:
                sColor = 'black'
                #pick color from colormap
            if sGeometry_type =='POINT':               
                aCoords_gcs = get_geometry_coordinates(pGeometry_in)
                aCoords_gcs = aCoords_gcs[:,0:2]
                #ax.plot(aCoords_gcs[0], aCoords_gcs[1], 'o', color= sColor, markersize=2, transform=ccrs.Geodetic())   
                aColor.append(sColor)    
                aPoint.append(aCoords_gcs)         
            else:
                if sGeometry_type =='LINESTRING':                    
                    aCoords_gcs = get_geometry_coordinates(pGeometry_in)
                    aCoords_gcs = aCoords_gcs[:,0:2]
                    nvertex = len(aCoords_gcs)
                    codes = np.full(nvertex, mpath.Path.LINETO, dtype=int )
                    codes[0] = mpl.path.Path.MOVETO
                    path = mpl.path.Path(aCoords_gcs, codes)
                    x, y = zip(*path.vertices)                    
                    aThickness.append(iThickness)
                    aColor.append(sColor)
                    aPolyline.append(list(zip(x, y)))
                else:
                    if sGeometry_type == 'POLYGON':                         
                        aCoords_gcs = get_geometry_coordinates(pGeometry_in)
                        if iFlag_fill == 1:                            
                            aColor.append(sColor)                            
                        else:                            
                            aColor.append('none')
                        aPolygon.append(aCoords_gcs[:, 0:2])                          
                    else:
                        if sGeometry_type == 'MULTIPOLYGON':     
                            for j in range(pGeometry_in.GetGeometryCount()):  
                                pPolygon = pGeometry_in.GetGeometryRef(j)               
                                aCoords_gcs = get_geometry_coordinates(pPolygon) 
                                if iFlag_fill == 1:
                                    aColor.append(sColor)
                                else:
                                    aColor.append('none')
                                aPolygon.append(aCoords_gcs[:, 0:2])  
                
            
            lID = lID + 1
        
        if len(aPoint) > 0:
            paths = [Path([point]) for point in aPoint]
            pPC = PathCollection(paths, alpha=0.8, edgecolor=aColor, 
                                 facecolor=aColor, linewidths=aThickness, transform=cpl.crs.Geodetic())
            ax.add_collection(pPC)     
        
        if len(aPolyline) > 0:
            #polyline
            pLC = LineCollection(aPolyline,  alpha=0.8, edgecolor=aColor,
                         facecolor='none', linewidths=aThickness, transform=cpl.crs.Geodetic())
            ax.add_collection(pLC)

        if len(aPolygon) > 0:
            aPatch = [Polygon(poly, closed=True) for poly in aPolygon]
            pPC = PatchCollection(aPatch, cmap=cmap, alpha=0.8, edgecolor=None, 
                                      facecolor=aColor, linewidths=0.25, 
                                      transform=cpl.crs.Geodetic())
            ax.add_collection(pPC)

    
    #reset extent
    ax.set_extent( aExtent )     
    ax.coastlines(color='black', linewidth=0.5)    
    iFlag_label = 0
    if iFlag_label == 1:
        sText = 'Manaus'
        dLongitude_label = -60.016667
        dLatitude_label  = -3.1
        ax.text(dLongitude_label, dLatitude_label, sText, 
                verticalalignment='center', horizontalalignment='center',                
                color='black', fontsize=iFont_size,transform=ccrs.Geodetic())

    if aLegend_in is not None:
        nlegend = len(aLegend_in)
        dLocation0 = 0.96
        for i in range(nlegend):
            sText = aLegend_in[i]
            #dLocation = 0.06 + i * 0.04
            dLocation = dLocation0 - i * 0.06
            ax.text(0.03, dLocation, sText,
                    verticalalignment='top', horizontalalignment='left',
                    transform=ax.transAxes,
                    color='black', fontsize=iFont_size)


    if iFlag_colorbar ==1:
        #ax_cb= fig.add_axes([0.75, 0.15, 0.02, 0.6])
        fig.canvas.draw()        
        ax_pos = ax.get_position() # get the original position      
        ax_cb = fig.add_axes([ax_pos.x1+0.06, ax_pos.y0, 0.02, ax_pos.height])   
        if iFlag_scientific_notation_colorbar==1:
            formatter = OOMFormatter(fformat= "%1.1e")
            cb = mpl.colorbar.ColorbarBase(ax_cb, orientation='vertical',
                                           cmap=cmap,
                                           norm=mpl.colors.Normalize(dValue_min, dValue_max),  # vmax and vmin
                                           extend=sExtend, format=formatter)
        else:
            formatter = OOMFormatter(fformat= "%1.1f")
            cb = mpl.colorbar.ColorbarBase(ax_cb, orientation='vertical',
                                           cmap=cmap,
                                           norm=mpl.colors.Normalize(dValue_min, dValue_max),  # vmax and vmin
                                           extend=sExtend, format=formatter)

        cb.ax.get_yaxis().set_ticks_position('right')
        cb.ax.get_yaxis().labelpad = 10
        cb.ax.set_ylabel(sUnit, rotation=270)
        cb.ax.tick_params(labelsize=6)

    gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True,
                      linewidth=1, color='gray', alpha=0.5, linestyle='--')
    gl.xformatter = LONGITUDE_FORMATTER
    gl.yformatter = LATITUDE_FORMATTER
    gl.xlabel_style = {'size': 10, 'color': 'k', 'rotation':0, 'ha':'right'}
    gl.ylabel_style = {'size': 10, 'color': 'k', 'rotation':90,'weight': 'normal'}  
    if iFlag_title==1:
        ax.set_title( sTitle )

    pDataset = pLayer = pFeature  = None

    if sFilename_output_in is None:
        plt.show()
    else:
        sDirname = os.path.dirname(sFilename_output_in)
        sFilename = os.path.basename(sFilename_output_in)
        sFilename_out = os.path.join(sDirname, sFilename)
        sExtension = os.path.splitext(sFilename)[1]
        if sExtension == '.png':
            plt.savefig(sFilename_out, bbox_inches='tight')
        else:
            if sExtension == '.pdf':
                plt.savefig(sFilename_out, bbox_inches='tight')
            else:
                plt.savefig(sFilename_out, bbox_inches='tight', format ='ps')
    
    #clean cache
    plt.close('all')
    plt.clf()



import os
import numpy as np
from numpy.random import RandomState
from osgeo import osr, gdal, ogr
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon
import cartopy as cpl
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
from pyearth.gis.location.get_geometry_coordinates import get_geometry_coordinates
from pyearth.visual.formatter import OOMFormatter
pProjection = cpl.crs.PlateCarree()  # for latlon data only

def map_vector_polygon_data(iFiletype_in,
                            sFilename_in,
                            sVariable_in=None,
                            sFilename_output_in=None,
                            iFlag_scientific_notation_colorbar_in=None,
                            iFlag_color_in = None,
                            iFlag_colorbar_in=None,
                            iFont_size_in=None,
                            iFlag_discrete_in=None,
                            sColormap_in=None,
                            sTitle_in=None,
                            iDPI_in=None,
                            iSize_x_in=None,
                            iSize_y_in=None,
                            dMissing_value_in=None,
                            dData_max_in=None,
                            dData_min_in=None,
                            sExtend_in=None,
                            sFont_in=None,
                            sUnit_in=None,
                            aLegend_in=None,
                            aExtent_in=None,
                            pProjection_map_in=None):
    """
    plot vector data on a map
    currently only support geojson and shapefile
    by default, the program will plot all the polygons in the file
    in the furture, the program will support to plot only a subset of polygons

    Args:
        iFiletype_in (_type_): _description_
        sFilename_in (_type_): _description_
        sFilename_output_in (_type_): _description_
        iFlag_scientific_notation_colorbar_in (_type_, optional): _description_. Defaults to None.
        sColormap_in (_type_, optional): _description_. Defaults to None.
        sTitle_in (_type_, optional): _description_. Defaults to None.
        iDPI_in (_type_, optional): _description_. Defaults to None.
        dMissing_value_in (_type_, optional): _description_. Defaults to None.
        dData_max_in (_type_, optional): _description_. Defaults to None.
        dData_min_in (_type_, optional): _description_. Defaults to None.
        sExtend_in (_type_, optional): _description_. Defaults to None.
        sUnit_in (_type_, optional): _description_. Defaults to None.
        aLegend_in (_type_, optional): _description_. Defaults to None.
    """

    if iFiletype_in == 1:  # geojson
        pDriver = ogr.GetDriverByName('GeoJSON')
    else:
        if iFiletype_in == 2:  # shapefile
            pDriver = ogr.GetDriverByName('Esri Shapefile')
        else:
            pDriver = ogr.GetDriverByName('Parquet')

    if pDriver is None:
        print('Driver not available')
        return
    
    if os.path.exists(sFilename_in) is False:
        print('File does not exist')
        return            

    pDataset = pDriver.Open(sFilename_in, gdal.GA_ReadOnly)
    pLayer = pDataset.GetLayer(0)

    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 8

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 8

    if iFont_size_in is not None:
        iFont_size = iFont_size_in
    else:
        iFont_size = 12

    if iFlag_colorbar_in is not None:
        iFlag_colorbar = iFlag_colorbar_in
    else:
        iFlag_colorbar = 0

    if dMissing_value_in is not None:
        dMissing_value = dMissing_value_in
    else:
        dMissing_value = -9999

    if dData_min_in is not None:
        iFlag_data_min = 1
        dData_min = dData_min_in
    else:
        iFlag_data_min = 0
        pass

    if dData_max_in is not None:
        iFlag_data_max = 1
        dData_max = dData_max_in
    else:
        iFlag_data_max = 0
        pass

    if iFlag_discrete_in is not None:
        iFlag_discrete = iFlag_discrete_in
    else:
        iFlag_discrete = 0

    if iFlag_scientific_notation_colorbar_in is not None:
        iFlag_scientific_notation_colorbar = iFlag_scientific_notation_colorbar_in
    else:
        iFlag_scientific_notation_colorbar = 0

    if sColormap_in is not None:
        sColormap = sColormap_in
    else:
        sColormap = 'rainbow'

    if sTitle_in is not None:
        sTitle = sTitle_in
        iFlag_title = 1
    else:
        iFlag_title = 0
        sTitle = ''

    if sExtend_in is not None:
        sExtend = sExtend_in
    else:
        sExtend = 'max'

    if sUnit_in is not None:
        sUnit = sUnit_in
    else:
        sUnit = ''

    if sFont_in is not None:
        sFont = sFont_in
    else:
        sFont = "Times New Roman"

    plt.rcParams["font.family"] = sFont
    plt.rcParams["mathtext.fontset"] = 'dejavuserif'

    if sVariable_in is not None:
        sVariable = sVariable_in
    else:
        sVariable = 'id'

    cmap = mpl.cm.get_cmap(sColormap)

    fig = plt.figure(dpi=iDPI)

    fig.set_figwidth(iSize_x)
    fig.set_figheight(iSize_y)

    pLayer = pDataset.GetLayer(0)
    pSrs = osr.SpatialReference()
    pSrs.ImportFromEPSG(4326)    # WGS84 lat/lon
    dLat_min = 90
    dLat_max = -90
    dLon_min = 180
    dLon_max = -180
    aValue = list()
    for pFeature in pLayer:
        pGeometry_in = pFeature.GetGeometryRef()
        sGeometry_type = pGeometry_in.GetGeometryName()
        dValue = float(pFeature.GetField(sVariable))
        aValue.append(dValue)
        if sGeometry_type == 'MULTIPOLYGON':        
            for i in range(pGeometry_in.GetGeometryCount()):                     
                pPolygon = pGeometry_in.GetGeometryRef(i)    
                aCoords_gcs = get_geometry_coordinates(pPolygon)   
                aCoords_gcs = np.array(aCoords_gcs)
                dLon_max = np.max([dLon_max, np.max(aCoords_gcs[:, 0])])
                dLon_min = np.min([dLon_min, np.min(aCoords_gcs[:, 0])])
                dLat_max = np.max([dLat_max, np.max(aCoords_gcs[:, 1])])
                dLat_min = np.min([dLat_min, np.min(aCoords_gcs[:, 1])])
        else:
            if sGeometry_type == 'POLYGON':    
                aCoords_gcs = get_geometry_coordinates(pGeometry_in)            
                aCoords_gcs = np.array(aCoords_gcs)
                dLon_max = np.max([dLon_max, np.max(aCoords_gcs[:, 0])])
                dLon_min = np.min([dLon_min, np.min(aCoords_gcs[:, 0])])
                dLat_max = np.max([dLat_max, np.max(aCoords_gcs[:, 1])])
                dLat_min = np.min([dLat_min, np.min(aCoords_gcs[:, 1])])


    aValue = np.array(aValue)
    if iFlag_discrete ==1:  
        #convert to integer
        aValue = np.array(aValue, dtype=int)
        #reorder it 
        aValue = np.unique(aValue)
        aValue = np.sort(aValue)
        nValue = len(aValue) #get unique values from the aValue

    if iFlag_data_min == 1:  # min is provided        
        dValue_min = dData_min  # np.min(aValue)
    else:
        aValue = aValue[aValue != dMissing_value]
        dValue_min = np.min(aValue)

    if iFlag_data_max == 1:  # max is provided        
        dValue_max = dData_max  # np.max(aValue)
    else:
        aValue = aValue[aValue != dMissing_value]
        dValue_max = np.max(aValue)
    
    aValue = np.clip(aValue, a_min=dData_min, a_max=dValue_max)

    print(dValue_min, dValue_max)       

    # print(sVariable,dValue_min, dValue_max )
    if dValue_max == dValue_min:
        iFlag_same_value = 1
        return
    else:
        iFlag_same_value = 0
        pass    

    if pProjection_map_in is not None:
        pProjection_map = pProjection_map_in
    else:
        pProjection_map = cpl.crs.Orthographic(central_longitude=0.50*(
            dLon_max+dLon_min),  central_latitude=0.50*(dLat_max+dLat_min), globe=None)

    ax = fig.add_axes([0.08, 0.1, 0.62, 0.7], projection=pProjection_map)
    ax.set_global()
    #use an advanced method for plotting
    aPolygon = list()
    aColor = list() 

    if iFlag_discrete ==1:                
        aIndex = np.linspace(0,1,nValue)
        prng = np.random.RandomState(1234567890)
        prng.shuffle(aIndex)
        #print(aIndex)
        colors = plt.cm.get_cmap(sColormap)(aIndex)
        cmap = ListedColormap(colors)
    else:
        cmap = mpl.cm.get_cmap(sColormap)    

    if aExtent_in is None:
        marginx = (dLon_max - dLon_min) / 20
        marginy = (dLat_max - dLat_min) / 20
        aExtent = [dLon_min - marginx, dLon_max + marginx,
                   dLat_min - marginy, dLat_max + marginy]
    else:
        aExtent = aExtent_in

    minx, miny, maxx, maxy = aExtent 
    pLayer.SetSpatialFilterRect(minx, miny, maxx, maxy)
    for pFeature in pLayer:
        pGeometry_in = pFeature.GetGeometryRef()
        sGeometry_type = pGeometry_in.GetGeometryName()
        # get attribute
        dValue = float(pFeature.GetField(sVariable))
    
        if dValue != dMissing_value:
            if dValue > dValue_max:
                dValue = dValue_max
                continue
            if dValue < dValue_min:
                dValue = dValue_min
                continue

            if iFlag_discrete ==1:     
                #use unique value method to assign color
                iValue = int(dValue)
                #find its index in the aValue array
                iColor_index = np.where(aValue == iValue)[0][0]                
            else:
                iColor_index = int((dValue - dValue_min) /
                               (dValue_max - dValue_min) * 255)
            # pick color from colormap            
            if sGeometry_type == 'MULTIPOLYGON':     
                for i in range(pGeometry_in.GetGeometryCount()):                     
                    pPolygon = pGeometry_in.GetGeometryRef(i)               
                    aCoords_gcs = get_geometry_coordinates(pPolygon)                
                    aCoords_gcs = np.array(aCoords_gcs)
                    aColor.append(cmap(iColor_index))
                    aPolygon.append(aCoords_gcs[:, 0:2])   
            else:
                if sGeometry_type == 'POLYGON': 
                    aCoords_gcs = get_geometry_coordinates(pGeometry_in)                
                    aCoords_gcs = np.array(aCoords_gcs)
                    aColor.append(cmap(iColor_index))
                    aPolygon.append(aCoords_gcs[:, 0:2]) 
                         
       
    aPatch = [Polygon(poly, closed=True) for poly in aPolygon]
    pPC = PatchCollection(aPatch, cmap=cmap, alpha=0.8, edgecolor=None, 
                                      facecolor=aColor, linewidths=0.25, 
                                      transform=cpl.crs.Geodetic())
    ax.add_collection(pPC)

    ax.set_extent(aExtent)
    ax.coastlines(color='black', linewidth=1)
    ax.set_title(sTitle)
    if aLegend_in is not None:
        nlegend = len(aLegend_in)
        dLocation0 = 0.96
        for i in range(nlegend):
            sText = aLegend_in[i]
            dLocation = dLocation0 - i * 0.06
            ax.text(0.03, dLocation, sText,
                    verticalalignment='top', horizontalalignment='left',
                    transform=ax.transAxes,
                    color='black', fontsize=iFont_size-2)


    if iFlag_colorbar == 1:
        fig.canvas.draw()
        # Section 2
        ax_pos = ax.get_position() # get the original position
        #use this ax to set the colorbar ax position
        ax_cb = fig.add_axes([ax_pos.x1+0.06, ax_pos.y0, 0.02, ax_pos.height])   
        if iFlag_scientific_notation_colorbar == 1:
            formatter = OOMFormatter(fformat="%1.1e")
            cb = mpl.colorbar.ColorbarBase(ax_cb, orientation='vertical',
                                           cmap=cmap,
                                           norm=mpl.colors.Normalize(
                                               dValue_min, dValue_max),  # vmax and vmin
                                           extend=sExtend, format=formatter)
        else:
            if iFlag_discrete ==1:            
                formatter = mpl.ticker.FuncFormatter(lambda x, pos: "{:.0f}".format(x))
                cb = mpl.colorbar.ColorbarBase(ax_cb, orientation='vertical',
                                           cmap=cmap,
                                           norm=mpl.colors.Normalize(
                                               dValue_min, dValue_max),  # vmax and vmin
                                           extend=sExtend, format=formatter)
            else:
                formatter = OOMFormatter(fformat="%1.2f")
                cb = mpl.colorbar.ColorbarBase(ax_cb, orientation='vertical',
                                           cmap=cmap,
                                           norm=mpl.colors.Normalize(
                                               dValue_min, dValue_max),  # vmax and vmin
                                           extend=sExtend, format=formatter)

        cb.ax.get_yaxis().set_ticks_position('right')
        cb.ax.get_yaxis().labelpad = 5
        cb.ax.set_ylabel(sUnit, rotation=90, fontsize=iFont_size-2)
        cb.ax.get_yaxis().set_label_position('left')
        cb.ax.tick_params(labelsize=iFont_size-2)

    gl = ax.gridlines(crs=cpl.crs.PlateCarree(), draw_labels=True,
                      linewidth=1, color='gray', alpha=0.5, linestyle='--')
    gl.xformatter = LONGITUDE_FORMATTER
    gl.yformatter = LATITUDE_FORMATTER

    gl.xlabel_style = {'size': 10, 'color': 'k', 'rotation': 0, 'ha': 'right'}
    gl.ylabel_style = {'size': 10, 'color': 'k',
                       'rotation': 90, 'weight': 'normal'}
    sDirname = os.path.dirname(sFilename_output_in)

    pDataset = pLayer = pFeature = None

    if sFilename_output_in is None:
        plt.show()
    else:
        sFilename = os.path.basename(sFilename_output_in)
        sFilename_out = os.path.join(sDirname, sFilename)
        sExtension = os.path.splitext(sFilename)[1]
        if sExtension == '.png':
            plt.savefig(sFilename_out, bbox_inches='tight')
        else:
            if sExtension == '.pdf':
                plt.savefig(sFilename_out, bbox_inches='tight')
            else:
                plt.savefig(sFilename_out, bbox_inches='tight', format='ps')
        plt.close('all')
        plt.clf()



import numpy as np
from pyearth.system.define_global_variables import *
def Google_MetersPerPixel( zoomLevel ):  
   
   # Return to the caller if there is an error.
   #On_Error, 2
   
   #; Need a zoom level?
   
   
   # Number of pixels in an image with a zoom level of 0.
   pixels_in_image = 256
   
   # The equitorial radius of the Earth assuming WGS-84 ellipsoid.
  
   
   # The number of meters per pixel.
   metersPerPixel = (2* np.pi * earth_radius) / pixels_in_image / np.power(2,zoomLevel)
   
   # Return the value.
   return metersPerPixel



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import cartopy as cpl

from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
from pyearth.visual.color.create_diverge_rgb_color_hex import create_diverge_rgb_color_hex
from pyearth.toolbox.data.cgpercentiles import cgpercentiles
from pyearth.visual.formatter import log_formatter
pProjection = cpl.crs.PlateCarree()



def map_raster_data_dc(aImage_in,
                       aImage_extent,
                       sFilename_output_in,
                       iFlag_scientific_notation_colorbar_in=None,
                       iFlag_contour_in=None,
                       sTitle_in=None,
                       aInterval_in=None,
                       aColor_in=None,
                       iDPI_in=None,
                       dMissing_value_in=None,
                       dData_max_in=None,
                       dData_min_in=None,
                       sExtend_in=None,
                       sFormat_contour_in=None,
                       sUnit_in=None,
                       aLabel_legend_in=None):

    aImage_in = np.array(aImage_in)

    pShape = aImage_in.shape
    nrow, ncolumn = aImage_in.shape
    iSize_x = ncolumn
    iSize_y = nrow
    sFilename_out = sFilename_output_in
    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_scientific_notation_colorbar_in is not None:
        iFlag_scientific_notation_colorbar = iFlag_scientific_notation_colorbar_in
    else:
        iFlag_scientific_notation_colorbar = 0

    if iFlag_contour_in is not None:
        iFlag_contour = iFlag_contour_in
    else:
        iFlag_contour = 0

    if sFormat_contour_in is not None:
        sFormat_contour = sFormat_contour_in
    else:
        sFormat_contour = '%1.1f'

    if aInterval_in is not None:
        iFlag_interval = 1
        aInterval = aInterval_in
        nInterval = len(aInterval)
    else:
        aPercentiles_in = np.arange(33, 67, 33)
        aInterval = cgpercentiles(
            aImage_in, aPercentiles_in, missing_value_in=dMissing_value_in)
        iFlag_interval = 0

    if aColor_in is not None:
        iFlag_colar = 1
        aColor = aColor_in
        ncolor = len(aColor)
        if iFlag_interval == 1:
            ncolor = nInterval + 1
        else:
            pass
    else:
        iFlag_colar = 0
        if iFlag_interval == 1:
            ncolor = nInterval + 1
            aColor = create_diverge_rgb_color_hex(ncolor)
        else:

            pass

    if dMissing_value_in is not None:
        dMissing_value = dMissing_value_in
    else:
        dMissing_value = np.nanmin(aImage_in)

    dummy_index = np.where(aImage_in == dMissing_value)
    aImage_in[dummy_index] = np.nan

    if dData_max_in is not None:
        dData_max = dData_max_in
    else:
        dData_max = np.nanmax(aImage_in)

    if dData_min_in is not None:
        dData_min = dData_min_in
    else:
        dData_min = np.nanmin(aImage_in)

    if sTitle_in is not None:
        sTitle = sTitle_in
        iFlag_title = 1
    else:
        iFlag_title = 0
        sTitle = ''

    if sExtend_in is not None:
        sExtend = sExtend_in
    else:
        sExtend = 'max'

    if sUnit_in is not None:
        sUnit = sUnit_in
    else:
        sUnit = ''

    dummy_index = np.where(aImage_in > dData_max)
    aImage_in[dummy_index] = dData_max

    dummy_index = np.where(aImage_in < dData_min)
    aImage_in[dummy_index] = dData_min

    fig = plt.figure(dpi=iDPI)

    ax = fig.add_axes([0.1, 0.1, 0.63, 0.7], projection=pProjection)

    # set a margin around the data
    ax.set_xmargin(0.05)
    ax.set_ymargin(0.10)

    if iFlag_contour == 1:
        aPercentiles_in = np.arange(33, 67, 33)
        levels = cgpercentiles(aImage_in, aPercentiles_in,
                               missing_value_in=dMissing_value)
        contourplot = ax.contour(aImage_in, levels, colors='k', origin='upper',
                                 extent=aImage_extent, transform=pProjection, linewidths=0.5)

        if iFlag_scientific_notation_colorbar == 1:
            ax.clabel(contourplot, contourplot.levels,
                      inline=True, fmt=log_formatter, fontsize=7)
        else:
            ax.clabel(contourplot, contourplot.levels,
                      inline=True, fmt=sFormat_contour, fontsize=7)

    aPseudo_image = np.full((nrow, ncolumn), fill_value=np.nan)

    dummy_index = np.where(aImage_in <= aInterval[0])
    if len(dummy_index) != 0:
        aPseudo_image[dummy_index] = 0

    for i in range(0, nInterval-1, 1):
        dIntervel0 = aInterval[i]
        dIntervel1 = aInterval[i+1]
        dummy_index = np.where((aImage_in > dIntervel0)
                               & (aImage_in <= dIntervel1))
        if len(dummy_index) != 0:
            aPseudo_image[dummy_index] = i+1

    dummy_index = np.where(aImage_in > aInterval[nInterval-1])
    if len(dummy_index) != 0:
        aPseudo_image[dummy_index] = nInterval

    cmap = (mpl.colors.ListedColormap(aColor[1:ncolor-1])
            .with_extremes(over=aColor[ncolor-1], under=aColor[0]))

    uni = np.unique(aPseudo_image[np.where(np.isfinite(aPseudo_image))])
    uni_index = np.array(uni).astype(int)
    dummy_color = list()
    for i in uni_index:
        dummy_color.append(aColor[i])
    cmap0 = mpl.colors.ListedColormap(dummy_color)

    dcrasterplot = ax.imshow(aPseudo_image, origin='upper',
                             extent=aImage_extent,
                             cmap=cmap0,
                             transform=pProjection)

    if aLabel_legend_in is not None:
        # plot the first on the top
        sText = aLabel_legend_in[0]
        dLocation = 0.96
        ax.text(0.03, dLocation, sText,
                verticalalignment='top', horizontalalignment='left',
                transform=ax.transAxes,
                color='black', fontsize=10)
        # plot the remaining on the bot
        nlegend = len(aLabel_legend_in)
        for i in range(1, nlegend, 1):
            sText = aLabel_legend_in[i]
            dLocation = nlegend * 0.06 - i * 0.05 - 0.03
            ax.text(0.03, dLocation, sText,
                    verticalalignment='top', horizontalalignment='left',
                    transform=ax.transAxes,
                    color='black', fontsize=10)

            pass

    ax.coastlines(color='black', linewidth=1)
    ax.set_title(sTitle)
    ax.set_extent(aImage_extent)

    gl = ax.gridlines(crs=cpl.crs.PlateCarree(), draw_labels=True,
                      linewidth=1, color='gray', alpha=0.5, linestyle='--')
    gl.xformatter = LONGITUDE_FORMATTER
    gl.yformatter = LATITUDE_FORMATTER

    gl.xlabel_style = {'size': 10, 'color': 'k', 'rotation': 0, 'ha': 'right'}
    gl.ylabel_style = {'size': 10, 'color': 'k',
                       'rotation': 90, 'weight': 'normal'}

    ax_cb = fig.add_axes([0.75, 0.1, 0.02, 0.7])
    bounds = np.linspace(0, nInterval-1, nInterval, endpoint=True)

    norm = mpl.colors.BoundaryNorm(bounds, cmap.N)
    cb = plt.colorbar(mpl.cm.ScalarMappable(cmap=cmap, norm=norm), cax=ax_cb,
                      extend=sExtend, extendfrac='auto',
                      ticks=bounds)

    cb.ax.get_yaxis().set_ticks_position('right')
    cb.ax.get_yaxis().labelpad = 10
    cb.ax.set_ylabel(sUnit, rotation=270)
    cb.ax.tick_params(labelsize=6)
    aLabel = list()
    for i in range(nInterval):
        if iFlag_scientific_notation_colorbar == 1:
            sLabel = '{:.1e}'.format(aInterval[i])
        else:
            sLabel = '{:.1f}'.format(aInterval[i])
        aLabel.append(sLabel)

    bounds0 = np.linspace(0, nInterval-1, nInterval)
    bounds = bounds0[0:nInterval]
    cb.ax.set_yticks(bounds)
    cb.ax.set_yticklabels(aLabel)
    # cb.ax.set_yticklabels([aLabel[int(i)] for i in bounds]) # add the labels

    plt.savefig(sFilename_out, bbox_inches='tight')
    plt.close('all')
    plt.clf()



import os
from pyearth.system.define_global_variables import *

def add_variable_to_netcdf(sFilename_old, sFilename_new, aData_in, sVariable_in, sUnit_in, aDimension_in):

    try:
        import netCDF4 as nc
    except ImportError as e:
        raise ImportError("The package 'netCDF4' is required for this function to run.") from e
    
    if os.path.exists(sFilename_old):
        print("Yep, I can read that file!")
    else:
        print("Nope, the path doesn't reach your file. Go research filepath in python")
        exit

    pDatasets_in = nc.Dataset(sFilename_old)
    netcdf_format = pDatasets_in.file_format
    #output file
    pDatasets_out = nc.Dataset(sFilename_new, "w", format=netcdf_format)
    aDimension_key=list()
    aDimension_value=list()
    for sKey, iValue in pDatasets_in.dimensions.items():
        dummy = len(iValue)
        if not iValue.isunlimited():            
            aDimension_key.append(sKey)
            aDimension_value.append(sKey)
            pDatasets_out.createDimension(sKey, dummy)            
        else:
            pDatasets_out.createDimension(sKey, dummy )

    #
    aDimension_list=list()
    for d in aDimension_in:
        index = aDimension_value.index(d)
        aDimension_list.append(aDimension_key[index])

    aDimension_tuple = tuple(aDimension_list)

    #Copy variables
    for sKey, aValue in pDatasets_in.variables.items():        
        
        # we need to take care of rec dimension
        dummy = aValue.dimensions
        outVar = pDatasets_out.createVariable(sKey, aValue.datatype, dummy )        
        for sAttribute in aValue.ncattrs():            
            outVar.setncatts( { sAttribute: aValue.getncattr(sAttribute) } )

        outVar[:] = aValue[:]
        # close the output file
    #add new variable 


    pVar3 = pDatasets_out.createVariable(sVariable_in, 'f4', aDimension_tuple  ) 
    pVar3[:] = aData_in
    pVar3.description = sVariable_in
    pVar3.unit = sUnit_in
    pDatasets_out.close()

    return


def add_multiple_variable_to_netcdf(sFilename_old, sFilename_new, aData_in, aVariable_in, aUnit_in, aDimension_in):
    try:
        import netCDF4 as nc
    except ImportError as e:
        raise ImportError("The package 'netCDF4' is required for this function to run.") from e
    
    if os.path.exists(sFilename_old):
        print("Yep, I can read that file!")
    else:
        print("Nope, the path doesn't reach your file. Go research filepath in python")
        exit
    missing_value=-9999
    pDatasets_in = nc.Dataset(sFilename_old)
    netcdf_format = pDatasets_in.file_format
    #output file
    pDatasets_out = nc.Dataset(sFilename_new, "w", format=netcdf_format)
    aDimension_key=list()
    aDimension_value=list()
    for sKey, iValue in pDatasets_in.dimensions.items():
        dummy = len(iValue)
        if not iValue.isunlimited():            
            aDimension_key.append(sKey)
            aDimension_value.append(dummy)
            pDatasets_out.createDimension(sKey, dummy)            
        else:
            pDatasets_out.createDimension(sKey, dummy )

    

    #Copy variables
    for sKey, aValue in pDatasets_in.variables.items():      
        # we need to take care of rec dimension
        dummy = aValue.dimensions
        outVar = pDatasets_out.createVariable(sKey, aValue.datatype, dummy )        
        for sAttribute in aValue.ncattrs():            
            outVar.setncatts( { sAttribute: aValue.getncattr(sAttribute) } )

        outVar[:] = aValue[:]
        # close the output file
    #add new variable 

    nVariable = len(aData_in)
    for iVariable in range(nVariable):
        aDimension_variable = aDimension_in[iVariable]
        sVariable = aVariable_in[iVariable]
        sUnit=aUnit_in[iVariable]
        aData = aData_in[iVariable]
        aDimension_list=list()
        for d in aDimension_variable:
            index = aDimension_value.index(d)
            aDimension_list.append(aDimension_key[index])

        aDimension_tuple = tuple(aDimension_list)

        pVar3 = pDatasets_out.createVariable(sVariable, 'f4', aDimension_tuple,fill_value=missing_value  ) 
        pVar3[:] = aData
        pVar3.description = sVariable
        pVar3.unit = sUnit
        
    pDatasets_out.close()

    return


import os
import numpy as np
from osgeo import  osr, gdal, ogr
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.collections import LineCollection

import cartopy as cpl
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER

from pyearth.gis.location.get_geometry_coordinates import get_geometry_coordinates

from pyearth.toolbox.math.stat.remap import remap


def map_vector_polyline_data(iFiletype_in,
                             sFilename_in,
                             sFilename_output_in,
                             iFlag_thickness_in =None,
                             sField_thickness_in=None,    
                             iFlag_color_in= None,    
                             iFlag_scientific_notation_colorbar_in=None,
                             iFlag_label_in = None,
                             sColormap_in = None,
                             sTitle_in = None,
                             iDPI_in = None,
                             iSize_x_in = None, 
                             iSize_y_in = None, 
                             iFont_size_in = None,
                             dMissing_value_in=None,
                             dData_max_in = None,
                             dData_min_in = None,
                             sFont_in = None,
                             sExtend_in =None,
                             sUnit_in=None,
                             aLegend_in = None,
                             aExtent_in = None,
                             pProjection_map_in = None):
    """
    Map a vector polyline data

    Args:
        iFiletype_in (int): _description_
        sFilename_in (_type_): _description_
        sFilename_output_in (_type_): _description_
        iFlag_thickness_in (_type_, optional): _description_. Defaults to None.
        sField_thickness_in (_type_, optional): _description_. Defaults to None.
        iFlag_scientific_notation_colorbar_in (_type_, optional): _description_. Defaults to None.
        sColormap_in (_type_, optional): _description_. Defaults to None.
        sTitle_in (_type_, optional): _description_. Defaults to None.
        iDPI_in (_type_, optional): _description_. Defaults to None.
        dMissing_value_in (_type_, optional): _description_. Defaults to None.
        dData_max_in (_type_, optional): _description_. Defaults to None.
        dData_min_in (_type_, optional): _description_. Defaults to None.
        sFont_in (_type_, optional): _description_. Defaults to None.
        sExtend_in (_type_, optional): _description_. Defaults to None.
        sUnit_in (_type_, optional): _description_. Defaults to None.
        aLegend_in (_type_, optional): _description_. Defaults to None.
        aExtent_in (_type_, optional): _description_. Defaults to None.
        pProjection_map_in (_type_, optional): _description_. Defaults to None.
    """



    sFilename_out= sFilename_output_in
    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300
    
    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 8

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 8

    if iFont_size_in is not None:
        iFont_size = iFont_size_in
    else:
        iFont_size = 12

    if iFlag_scientific_notation_colorbar_in is not None:
        iFlag_scientific_notation_colorbar = iFlag_scientific_notation_colorbar_in
    else:
        iFlag_scientific_notation_colorbar = 0

    if iFlag_thickness_in is not None:
        iFlag_thickness = iFlag_thickness_in
    else:
        iFlag_thickness = 0

    if sField_thickness_in is not None:
        sField_thickness = sField_thickness_in
    else:
        sField_thickness = ''

    if iFlag_color_in is not None:
        iFlag_color = iFlag_color_in
    else:
        iFlag_color = 0


    if iFiletype_in == 1: #geojson
        pDriver = ogr.GetDriverByName('GeoJSON')
    else:
        if iFiletype_in == 2: #shapefile
            pDriver = ogr.GetDriverByName('Esri Shapefile')

    pDataset = pDriver.Open(sFilename_in, gdal.GA_ReadOnly)
    pLayer = pDataset.GetLayer(0)


    if sColormap_in is not None:
        sColormap = sColormap_in
    else:
        sColormap =  'Spectral'

    if sTitle_in is not None:
        sTitle = sTitle_in
        iFlag_title =1
    else:
        iFlag_title=0
        sTitle =  ''

    if sExtend_in is not None:
        sExtend = sExtend_in
    else:
        sExtend =  'max'

    if sUnit_in is not None:
        sUnit = sUnit_in
    else:
        sUnit =  ''
    
    if sFont_in is not None:
        sFont = sFont_in
    else:    
        sFont = "Times New Roman"

    plt.rcParams["font.family"] = sFont
    
    cmap = mpl.cm.get_cmap(sColormap)  
 
    pSrs = osr.SpatialReference()
    pSrs.ImportFromEPSG(4326)    # WGS84 lat/lon

    lID = 0
    dLat_min = 90
    dLat_max = -90
    dLon_min = 180
    dLon_max = -180
    for pFeature in pLayer:
        pGeometry_in = pFeature.GetGeometryRef()
        sGeometry_type = pGeometry_in.GetGeometryName()        
        if sGeometry_type =='LINESTRING':            
            aCoords_gcs =   get_geometry_coordinates(pGeometry_in)            
            aCoords_gcs= np.array(aCoords_gcs)
            aCoords_gcs = aCoords_gcs[:,0:2]

            dLon_max = np.max( [dLon_max, np.max(aCoords_gcs[:,0])] )
            dLon_min = np.min( [dLon_min, np.min(aCoords_gcs[:,0])] )
            dLat_max = np.max( [dLat_max, np.max(aCoords_gcs[:,1])] )
            dLat_min = np.min( [dLat_min, np.min(aCoords_gcs[:,1])] )           
           
                    
    if pProjection_map_in is not None:
        pProjection_map = pProjection_map_in
    else:
        pProjection_map = cpl.crs.Orthographic(central_longitude =  0.50*(dLon_max+dLon_min),  central_latitude = 0.50*(dLat_max+dLat_min), globe=None)
   
    fig = plt.figure( dpi=iDPI)
    fig.set_figwidth( iSize_x )
    fig.set_figheight( iSize_y )
    ax = fig.add_axes([0.1, 0.15, 0.75, 0.8] , projection=pProjection_map ) #request.crs
    ax.set_global()

    n_colors = pLayer.GetFeatureCount()
    colours = mpl.cm.rainbow(np.linspace(0, 1, n_colors))
    if iFlag_thickness ==1:
        aField =list()
        for pFeature in pLayer:
            dField = pFeature.GetField(sField_thickness)
            aField.append(dField)           

        aField = np.array(aField)
        dField_max = np.max(aField)
        dField_min = np.min(aField)

    iThickness_max = 2.5
    iThickness_min = 0.3

    #switch to collection in next development
    aPolyline = list()
    aColor = list()
    aThickness = list()

    if aExtent_in is None:
        marginx  = (dLon_max - dLon_min) / 20
        marginy  = (dLat_max - dLat_min) / 20
        aExtent = [dLon_min - marginx , dLon_max + marginx , dLat_min - marginy , dLat_max + marginy]
    else:
        aExtent = aExtent_in

    ax.set_extent(aExtent)

    minx, miny, maxx, maxy = aExtent 
    pLayer.SetSpatialFilterRect(minx, miny, maxx, maxy)

    for pFeature in pLayer:
        pGeometry_in = pFeature.GetGeometryRef()
        sGeometry_type = pGeometry_in.GetGeometryName()
        if iFlag_thickness ==1:
            dField = pFeature.GetField(sField_thickness)
        if sGeometry_type =='LINESTRING':
            aCoords_gcs = get_geometry_coordinates(pGeometry_in)                
            aCoords_gcs = np.array(aCoords_gcs)
            aCoords_gcs = aCoords_gcs[:,0:2]
            nvertex = len(aCoords_gcs)        

            if nvertex == 2 :
                dLon_label = 0.5 * (aCoords_gcs[0][0] + aCoords_gcs[1][0] )
                dLat_label = 0.5 * (aCoords_gcs[0][1] + aCoords_gcs[1][1] )
            else:
                lIndex_mid = int(nvertex/2)
                dLon_label = aCoords_gcs[lIndex_mid][0]
                dLat_label = aCoords_gcs[lIndex_mid][1]

            codes = np.full(nvertex, mpl.path.Path.LINETO, dtype=int )
            codes[0] = mpl.path.Path.MOVETO
            path = mpl.path.Path(aCoords_gcs, codes)
            x, y = zip(*path.vertices)
            if iFlag_thickness ==1:
                iThickness = remap( dField, dField_min, dField_max, iThickness_min, iThickness_max )
            else:
                iThickness = 1.0

            aThickness.append(iThickness)

            if iFlag_color ==1:
                color_index = (dField-dField_min ) /(dField_max - dField_min )            
                rgba = cmap(color_index)
            else:
                rgba = 'blue'

            if sColormap_in is not None:
                #line, = ax.plot(x, y, color=rgba,linewidth=iThickness, transform=cpl.crs.Geodetic())
                aColor.append(rgba)
            else:
                if n_colors < 10:
                    #line, = ax.plot(x, y, color= colours[lID],linewidth=iThickness, transform=cpl.crs.Geodetic())
                    aColor.append(colours[lID])
                else:
                    #line, = ax.plot(x, y, color= 'black',linewidth=iThickness, transform=cpl.crs.Geodetic())
                    aColor.append('black')
            
            aPolyline.append(list(zip(x, y)))

            lID = lID + 1

    pLC = LineCollection(aPolyline,  alpha=0.8, edgecolor=aColor,
                         facecolor='none', linewidths=aThickness, transform=cpl.crs.Geodetic())
    ax.add_collection(pLC)
    
    ax.set_extent(aExtent)

    ax.coastlines(color='black', linewidth=1)

    gl = ax.gridlines(crs=cpl.crs.PlateCarree(), draw_labels=True,
                      linewidth=1, color='gray', alpha=0.5, linestyle='--')
    gl.xformatter = LONGITUDE_FORMATTER
    gl.yformatter = LATITUDE_FORMATTER
    gl.xlocator = mpl.ticker.MaxNLocator(5)
    gl.ylocator = mpl.ticker.MaxNLocator(5)
    gl.xlabel_style = {'size': 8, 'color': 'k', 'rotation':0, 'ha':'right'}
    gl.ylabel_style = {'size': 8, 'color': 'k', 'rotation':90,'weight': 'normal'}

    if aLegend_in is not None:
        nlegend = len(aLegend_in)
        dLocation0 = 0.96
        for i in range(nlegend):
            sText = aLegend_in[i]
            dLocation = dLocation0 - i * 0.06
            ax.text(0.03, dLocation, sText,
                    verticalalignment='top', horizontalalignment='left',
                    transform=ax.transAxes,
                    color='black', fontsize=iFont_size-2 )

            pass

    if iFlag_title is None:
        ax.set_title( sTitle )
    else:
        if iFlag_title==1:
            ax.set_title( sTitle )
        else:
            pass
        ax.set_title(sTitle)

    pDataset = pLayer = pFeature  = None   
    sDirname = os.path.dirname(sFilename_output_in)
    if sFilename_output_in is None:
        plt.show()
    else:
        sFilename_out = os.path.join(sDirname, sFilename_output_in)
        plt.savefig(sFilename_out, bbox_inches='tight')   
        plt.close('all')
        plt.clf()
    




import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

def histogram_w_cdf_plot(aData,
                         sFilename_out,
                         iSize_x_in=None,
                         iSize_y_in=None,
                         iDPI_in=None,
                         dMin_x_in=None,
                         dMax_x_in=None,
                         dSpace_x_in=None,
                         sLabel_x_in=None,
                         sLabel_y_in=None,
                         sTitle_in=None,
                         sLabel_legend_in=None):
    """
    Draw a histogram for single dataset
    """

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12
    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9
    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if dMin_x_in is not None:
        dMin_x = dMin_x_in
    else:
        dMin_x = np.min(aData)

    if dMax_x_in is not None:
        dMax_x = dMax_x_in
    else:
        dMax_x = np.max(aData)

    if dSpace_x_in is not None:
        dSpace_x = dSpace_x_in
    else:
        # it may be calculated
        pass

    if sLabel_x_in is not None:
        sLabel_x = sLabel_x_in
    else:
        sLabel_x = ''

    if sLabel_y_in is not None:
        sLabel_y = sLabel_y_in
    else:
        sLabel_y = ''

    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    if sLabel_legend_in is not None:
        sLabel_legend = sLabel_legend_in
    else:
        sLabel_legend = ''

    good_index = np.where((aData >= dMin_x) & (aData <= dMax_x))

    fig = plt.figure(dpi=iDPI)
    fig.set_figwidth(iSize_x)
    fig.set_figheight(iSize_y)

    left, width = 0.15, 0.7
    bottom, height = 0.1, 0.85
    spacing = 0.005
    rect_histogram = [left, bottom, width, height]

    ax_histo = plt.axes(rect_histogram)
    ax_histo.tick_params(direction='in', top=True, right=True)
    aData = aData[good_index]
    ax_histo.hist(aData, int((dMax_x-dMin_x)/dSpace_x), density=True)
    ax_histo.set_xlabel(sLabel_x, fontsize=13)
    ax_histo.set_ylabel(sLabel_y, fontsize=13)
    ax_histo.set_xlim(dMin_x, dMax_x)

    ax_histo.axis('on')
    ax_histo.grid(which='major', color='white', linestyle='-', axis='y')

    handles = [mpl.patches.Rectangle(
        (0, 0), 1, 1, fc="white", ec="white", lw=0, alpha=0)] * 1

    # create the corresponding number of labels (= the text you want to display)
    labels = []
    labels.append(sLabel_legend)
    # create the legend, supressing the blank space of the empty line symbol    and the
    # padding between symbol and label by setting handlelenght and  handletextpad

    ax_histo.legend(handles, labels, loc="upper right", fontsize=12,
                    fancybox=True, framealpha=0.7,
                    handlelength=0, handletextpad=0)

    ax_cdf = ax_histo.twinx()
    count, bins_count = np.histogram(aData, bins=100)
    # finding the PDF of the histogram using count values
    pdf = count / sum(count)
    cdf = np.cumsum(pdf)
    ax_cdf.plot(bins_count[1:], cdf)

    ax_histo.set_title(sTitle)
    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')




from pyearth.system.define_global_variables import *
def prepare_parafly_python_command_file(iIndex_start,
                                        iIndex_end,
                                        nThread,
                                        sFilename_parafly,
                                        sFilename_python):
    """
    Prepare a parafly file which runs python script

    Args:
        iIndex_start (int): The start index
        iIndex_end (int): The end index
        nThread (int): Number of thread
        sFilename_parafly (string): The python binary
        sFilename_python (string): The parafly filename
    """

    ofs = open(sFilename_parafly, "w")  # write mode
    nTask = iIndex_end - iIndex_start + 1
    nChunkPerThread = nTask // nThread

    for iRank in range(nThread):
        if iRank == 0:
            iStart = (nThread-1) * nChunkPerThread + iIndex_start
            iEnd = iIndex_end
        else:
            iStart = (iRank-1) * nChunkPerThread + iIndex_start
            iEnd = (iRank) * nChunkPerThread + iIndex_start - 1

        sStart = "{:0d}".format(iStart)
        sEnd = "{:0d}".format(iEnd)
        sLine = 'python ' + sFilename_python \
            + ' --iIndex_start ' + sStart \
            + ' --iIndex_end ' + sEnd + '\n'

        ofs.write(sLine)

    ofs.close()

    return




import statistics
import numpy as np

from pyearth.system.define_global_variables import *


def search_neighbors(iIndex_in,
                     aArray_in,
                     iWindow_size_in=None):
    """
    Gap filling a 2D array

    Args:
        iIndex_in (int): [description]
        aArray_in (numpy.array): Input data
        iWindow_size_in (int, optional): Search window. Defaults to None.

    Returns:
        numpy.array: Gap filled array
    """
    missing_value = np.nan
    if iWindow_size_in is not None:
        iWindow_size = iWindow_size_in
    else:
        iWindow_size = 1

    iLength = len(aArray_in)
    iStart = iWindow_size
    iEnd = iLength - iWindow_size
    iLength_out = iWindow_size * 2 + 1
    aArray_out = np.full(iLength_out, np.nan, dtype=float)
    if iIndex_in >= iStart and iIndex_in < iEnd:
        aArray_out = aArray_in[(iIndex_in - iWindow_size)
                                : (iIndex_in + iWindow_size+1)]

    return aArray_out


def gap_fill_by_window(aArray_in, iWindow_size_in=None):
    """
    Gap filling a 2D array

    Args:
        aArray_in (numpy.array): 2D array
        iWindow_size_in (int, optional): search window. Defaults to None.

    Returns:
        numpy.array: Gap filled 2D array
    """
    iLength = len(aArray_in)
    iWindow_size = iWindow_size_in
    aArray_out = aArray_in

    iStart = iWindow_size
    iEnd = iLength - iWindow_size

    for iIndex in range(iStart, iEnd):
        dummy = aArray_in[iIndex]
        if (np.isnan(dummy)):
            aNeighbors = search_neighbors(
                iIndex, aArray_out, iWindow_size_in=iWindow_size)

            aNeighbors = aNeighbors.reshape(len(aNeighbors))
            nan_index = np.where(np.isnan(aNeighbors) == True)

            nan_count = len(nan_index[0])
            # only if half of the aNeighbors are qualified till we use this
            # gap fill
            if nan_count <= (iWindow_size + 1):
               # aNeighbors=gap_fill(aNeighbors,/spline)

                good_index = np.where(np.isnan(aNeighbors) == False)

                aNeighbors_good = aNeighbors[good_index]
                good_count = len(aNeighbors_good)

                if good_count < 10:
                    aArray_out[iIndex] = statistics.median(aNeighbors_good)
                    pass

                else:
                    aArray_out[iIndex] = statistics.median(aNeighbors_good)
                    pass

            else:

                pass
             # good elements count is not enough

        else:

            pass

    return aArray_out



"""
this module will be used to define all the global variables 
"""

import platform  
from pathlib import Path
import getpass
sPlatform_os = platform.system()

sUsername = getpass.getuser()

sWorkspace_home = str(Path.home())

if sPlatform_os == 'Windows':  #windows
    slash = '\\'
    sMachine ='None'
       
    sWorkspace_scratch = 'C:'    
else:  #linux or unix
    slash = '/'    
   
    if (sPlatform_os == 'Linux'):
        sWorkspace_scratch =  sWorkspace_home + slash + 'scratch'            
        
    else:
        if (sPlatform_os == 'Darwin'):
            sMachine ='mac'
            sWorkspace_scratch =  sWorkspace_home + slash + 'scratch'  
        else:
            pass

#data file type
sExtension_txt = '.txt'
sExtension_envi = '.dat'
sExtension_tiff = '.tif'
sExtension_header ='.hdr'
sExtension_netcdf = '.nc'
sExtension_shapefile = '.shp'
sExtension_json = '.json'

#graphics format

sExtension_png = '.png'
sExtension_jpg = '.jpg'
sExtension_ps = '.ps'
sExtension_vtk = '.vtk'


#constant values
missing_value = -9999.0

nmonth = 12 #be careful with this one

#unit conversion

iMonth_start = 1
iMonth_end = 12
mms2mmd = 24 * 3600.0
feet2meter = 0.3048
inch2mm = 25.4
cms2cmd = 24 * 3600


earth_radius = 6378137.0



import numpy as np

def scatter_lowess(x, y, f=1./3.):
    """
    Basic LOWESS smoother with uncertainty.
    Note:
        - Not robust (so no iteration) and
             only normally distributed errors.
        - No higher order polynomials d=1
            so linear smoother.
    """
    # get some paras
    xwidth = f*(x.max()-x.min()) # effective width after reduction factor
    N = len(x) # number of obs
    # Don't assume the data is sorted
    order = np.argsort(x)
    # storage
    y_sm = np.zeros_like(y)
    y_stderr = np.zeros_like(y)
    # define the weigthing function -- clipping too!
    tricube = lambda d : np.clip((1- np.abs(d)**3)**3, 0, 1)
    # run the regression for each observation i
    for i in range(N):
        dist = np.abs((x[order][i]-x[order]))/xwidth
        w = tricube(dist)
        # form linear system with the weights
        A = np.stack([w, x[order]*w]).T
        b = w * y[order]
        ATA = A.T.dot(A)
        ATb = A.T.dot(b)
        # solve the syste
        sol = np.linalg.solve(ATA, ATb)
        # predict for the observation only
        yest = A[i].dot(sol)# equiv of A.dot(yest) just for k
        place = order[i]
        y_sm[place]=yest
        sigma2 = (np.sum((A.dot(sol) -y [order])**2)/N )
        # Calculate the standard error
        y_stderr[place] = np.sqrt(sigma2 *
                                  A[i].dot(np.linalg.inv(ATA)
                                  ).dot(A[i]))
    return y_sm, y_stderr, order
